from w.services.abstract_service import AbstractService


class ListService(AbstractService):
    @staticmethod
    def list_differences(list1, list2):
        """List elements from list1 not in list2 and elements from list2 not in list1"""
        # Convert lists to sets
        set1 = set(list1)
        set2 = set(list2)
        # Get the differences between two sets
        diff = (set1 - set2).union(set2 - set1)
        return list(diff)

    @classmethod
    def are_same(cls, list1, list2):
        """
        Check if elements from list1 are all in list2 or
        elements from list2 are all in list1
        """
        return cls.list_differences(list1, list2) == []

    @classmethod
    def are_different(cls, list1, list2):
        """
        Check if elements from list1 are not in list2 or
        elements from list2 are not in list1
        """
        return not cls.are_same(list1, list2)



import datetime
from decimal import Decimal

import pytest

from w.services.technical.filesystem_service import FilesystemService
from w.services.technical.json_service import JsonService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestJsonService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.no_dumpable_data = {
            "str": "string",
            "int": 15,
            "float": 18.2,
            "decimal": Decimal(11.115),
            "datetime": datetime.datetime(2018, 7, 7, 12, 15, 35),
            "date": datetime.date(2018, 7, 7),
            "bytes": b"a bytes text a translate",
            "tuple": ("string", Decimal(11.5)),
            "list": [datetime.datetime(2018, 7, 7, 12, 15, 35), Decimal(11.115)],
        }

        cls.dumpable_data = {
            "a value": "a data",
            "a list": [1, 12, 13, "string"],
            "a dict": {"key": "value", "key2": ["value", "valuebis"]},
        }
        cls.valid_json = (
            '{"menu":{"id":"file","value":"File","popup":{"menuitem":[{"value":"New",'
            '"onclick":"CreateNewDoc()"},{"value":"Open","onclick":"OpenDoc()"},'
            '{"value":"Close","onclick":"CloseDoc()"}]}}}'
        )

    def setup_method(self):
        JsonService.clear()
        self.clean_sandbox()

    """
    to_json_dumpable
    """

    def test_json_dumpable_with_success_return_dict(self):
        """ Ensure method succeed on dict  """
        self.assert_equals_resultset(
            JsonService.to_json_dumpable(self.no_dumpable_data)
        )

    """
    dump_to_file
    """

    def test_dump_to_file_with_success_return_none(self):
        """ Ensure method succeed """
        filename = self.get_sandbox_dir("simple_dump.json")
        assert JsonService.dump_to_file(self.dumpable_data, filename) is None
        self.assert_file_exists(filename)
        self.assert_equals_resultset(FilesystemService.read_file(filename))

    def test_dump_to_file_with_human_readable_return_none(self):
        """ Ensure method succeed with human readable option"""
        filename = self.get_sandbox_dir("simple_dump.json")
        assert (
            JsonService.dump_to_file(self.dumpable_data, filename, human_readable=True)
            is None
        )
        self.assert_file_exists(filename)
        self.assert_equals_resultset(FilesystemService.read_file(filename))

    """
    load_from_file
    """

    def test_load_with_file_not_exists_raise_runtime_error(self):
        """ Ensure load unknown file raise RuntimeError """
        with pytest.raises(RuntimeError, match="unknown.json does not exists"):
            JsonService.load_from_file("unknown.json")

    def test_load_from_file_with_success_return_dict(self):
        """ Ensure load json file succeed """
        filename = self.get_datasets_dir("json/example.json")
        self.assert_equals_resultset(JsonService.load_from_file(filename))

    """
    load_from_str
    """

    def test_load_from_str_with_success_return_dict(self):
        """ Ensure method succeed """
        self.assert_equals_resultset(JsonService.load_from_str(self.valid_json))

    """
    is_valid
    """

    def test_is_valid_with_no_valid_json_return_false(self):
        data = "]invalid json}"
        assert not JsonService.is_valid(data)

    def test_is_valid_with_success_return_true(self):
        assert JsonService.is_valid(self.valid_json)



from typing import Dict


class SerializerMixin:
    # serializer list
    _serializers: Dict = {}

    @classmethod
    def serialize(cls, serializer_key, queryset, many=False):
        if serializer_key not in cls._serializers:
            raise RuntimeError(
                f"{serializer_key} missing, add it to {__class__}._serializers"
            )
        serializer = cls._serializers[serializer_key](queryset, many=many)
        return serializer.data



from contextlib import contextmanager
from unittest.mock import patch

from w.services.abstract_service import AbstractService


def get_target(config):
    return config["service"].get_patch_target(config["method_name"])


@contextmanager
def mock_services(mock_configs):
    """
    Patch services methods calls

    Args:
        mock_configs(list): list of mock config
            config = [{
              "service": <service>
              "method_name": <method to patch>,
              "return_value": <return value>
              "side_effect": [<return value>]
            }]

    Returns:
        dict: {<method_name>: mocks, ...}
    """
    patchers = []
    mocks = {}
    for config in mock_configs:
        patcher = patch(get_target(config))
        mock = patcher.start()
        if "return_value" in config:
            mock.return_value = config["return_value"]
        if "side_effect" in config:
            mock.side_effect = config["side_effect"]
        mocks[config["method_name"]] = mock
        patchers.append(patcher)

    yield mocks

    for patcher in patchers:
        patcher.stop()


@contextmanager
def mock_service(
    service: AbstractService, method_name: str, return_value=None, side_effect=None
):
    """
    Patch a service method calls
    """
    params = {"target": get_target({"service": service, "method_name": method_name})}
    # not None for Boolean return
    if side_effect is not None:
        params["side_effect"] = side_effect
    else:
        params["return_value"] = return_value
    with patch(**params) as m:
        yield m



import factory
from django.contrib.auth import models

# crypted version of testtest
crypted = "argon2$argon2i$v=19$m=512,t=2,p=2$TXFvZTQ3RnVBQUV1$MIzwaXUe6yRwxm5zdTDmDg"


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.User

    username = factory.sequence(lambda n: f"user_{n}")
    first_name = factory.LazyAttribute(lambda o: f"{o.username[:20]}_firstname")
    last_name = factory.LazyAttribute(lambda o: f"{o.username}_lastname")
    email = factory.LazyAttribute(lambda o: f"{o.username}@test.com")

    password = crypted
    is_active = True

    is_staff = False
    is_superuser = False


class GroupFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Group

    name = factory.Sequence(lambda n: f"group_{n}")


class PermissionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Permission

    name = factory.Sequence(lambda n: f"permission_{n}")
    content_type = factory.Iterator(models.ContentType.objects.all())
    codename = factory.lazy_attribute(
        lambda o: f"{o.content_type.model}_{o.name.lower().replace(' ', '_')}"
    )



from w.tests.builders.abstract_test_builder import AbstractTestBuilder
from w.tests.fixtures.datasets.builders.factory_boy import (
    ExampleWithInternalDependenciesFactory,
    InternalDependencyOneFactory,
    InternalDependencyTwoFactory,
    RelatedDependencyFactory,
)


class BuilderWithRelatedDependencies(AbstractTestBuilder):
    _factories = {
        "example": ExampleWithInternalDependenciesFactory,
        "internal_one": InternalDependencyOneFactory,
        "internal_two": InternalDependencyTwoFactory,
        "related_dependency": RelatedDependencyFactory,
    }
    _main_factory_key = "example"

    def with_internal_one(self, **attrs):
        self._set_foreign_key_factory_attrs("internal_one", attrs)
        return self

    def with_internal_two(self, **attrs):
        self._set_foreign_key_factory_attrs("internal_two", attrs)
        return self

    def with_related_dependency(self, **attrs):
        self._add_related_factory_attrs("related_dependency", attrs)
        return self

    def with_related_dependencies(self, list_attrs: list):
        self._set_related_factory_attrs("related_dependency", list_attrs)
        return self

    def with_nb_related_dependencies(self, nb, **attrs):
        self._set_nb_related_factory_attrs("related_dependency", nb, attrs)
        return self

    def get_related_dependency_built(self):
        return self._get_related_built("related_dependency")



from rest_framework import routers

from w.tests.fixtures.datasets.django_app.viewsets import SimpleViewset, ModelViewset

router = routers.SimpleRouter()
router.register(r"simples", SimpleViewset, basename="simple")
router.register(r"models", ModelViewset, basename="model")
urlpatterns = router.urls



class SsoValidToken:
    def __init__(self, decoded_token):
        self.sso_uuid = decoded_token["sub"]
        self.first_name = decoded_token["given_name"]
        self.last_name = decoded_token["family_name"]
        self.username = decoded_token["username"]
        self.email = decoded_token["email"]
        self.list_apps = list(decoded_token["resource_access"].keys())

    def to_dict(self):
        return {
            "sso_uuid": self.sso_uuid,
            "first_name": self.first_name,
            "last_name": self.last_name,
            "username": self.username,
            "email": self.email,
            "list_apps": self.list_apps,
        }



from rest_framework import viewsets, status
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from w import exceptions
from w.services.abstract_model_service import AbstractModelService


class ViewSet(viewsets.ViewSet):
    serializers = {"default": None}

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_key = kwargs.pop("serializer_key", None)
        serializer_class = self.get_serializer_class(serializer_key)
        return serializer_class(*args, **kwargs)

    @classmethod
    def get_serializer_by_key(cls, key):
        """
        Args:
            key: serializer key

        Returns:
            Serializer
        """
        return cls.serializers.get(key, cls.serializers["default"])

    def get_serializer_class(self, serializer_key=None):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.

        You may want to override this if you need to provide different
        serializations depending on the incoming request.

        (Eg. admins get full serialization, others get basic serialization)

        Returns:
            Serializer: Drf or Serpy
        """
        serializer_key = serializer_key if serializer_key else self.action  # noqa
        return self.get_serializer_by_key(serializer_key)

    def get_action(self):
        # noinspection PyUnresolvedReferences
        return self.action_map.get(self.request.method.lower())

    def get_response(self, data, **params) -> Response:
        """
        Get response from data.

        Automatically serialize data from action name or parameter serializer_key

        Args:
            data(mixed): data to return in response
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_200_OK

        Returns:
            Response
        """
        default = {
            "serializer_key": None,
            "status_code": status.HTTP_200_OK,
        }
        params = {**default, **params}
        serializer_class = self.get_serializer_class(params["serializer_key"])
        many = isinstance(data, list)
        serializer = serializer_class(data, many=many)

        return Response(serializer.data, status=params.get("status_code"))

    def get_post_response(self, data, **params) -> Response:
        """
        Get post response from data.

        Automatically serialize data from action name or parameter serializer_key

        Args:
            data(mixed): data to return in response
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_201_CREATED
        Returns:
            Response
        """
        default = {
            "serializer_key": None,
            "status_code": status.HTTP_201_CREATED,
        }
        params = {**default, **params}
        return self.get_response(data, **params)

    def get_validated_data(self, serializer_key=None):
        """
        Validate request data (throw exception if invalid) and return validated data

        Args:
            serializer_key(None|str): serializer to use,
                default is "<action name>_validation"

        Returns:
            dict
        """
        if serializer_key is None:
            serializer_key = f"{self.action}_validation"
        if serializer_key not in self.serializers:
            raise RuntimeError(f"You need to define '{serializer_key}' serializer")
        serializer_class = self.get_serializer_class(serializer_key)
        serializer = serializer_class(data=self.request.data)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data

    @classmethod
    def _simplify_errors(cls, errors):
        if isinstance(errors, dict):
            return {attr: cls._simplify_errors(errs) for attr, errs in errors.items()}
        if isinstance(errors, list) and isinstance(errors[0], dict):
            # Simplify many to many relations errors
            many2many_errors = []
            for item_errors in errors:
                many2many_errors.append(cls._simplify_errors(item_errors))
            return many2many_errors
        return str(errors[0])

    @staticmethod
    def get_error_validation_response(error: ValidationError):
        """
        Build validation error response with status code of 422 and
        by keeping only first validation message
        """
        data = {}
        for attr, errors in error.detail.items():
            data[attr] = str(errors[0])
        return Response(data, status=status.HTTP_422_UNPROCESSABLE_ENTITY)

    def check_is_valid(self, serializer_key=None):
        try:
            validated_data = self.get_validated_data(serializer_key)
        except ValidationError as e:
            errors = self._simplify_errors(e.detail)
            raise exceptions.ValidationError(errors)
        return dict(validated_data)

    @action(methods=["post"], detail=False)
    def validate(self, request):
        """
        Validate using query params action to retrieve serializer name

        POST <viewset>/validate/?action=<action name>
        """
        action_name = self.request.GET.get("action")
        if action_name is None:
            raise RuntimeError("missing query parameter 'action'")
        validated_data = self.check_is_valid(serializer_key=f"{action_name}_validation")
        return Response(validated_data)


class ModelServiceViewSet(ViewSet):
    service: AbstractModelService

    def get_detail_response(
        self, instance=None, serializer_key=None, status_code=status.HTTP_200_OK
    ) -> Response:
        """
        Build response

        Args:
            instance(Model|None): if None use service.get_by_pk
            serializer_key: default: None,
            status_code: default: status.HTTP_200_OK

        Returns:

        """
        if instance is None:
            instance = self.service.check_by_pk(self.kwargs["pk"])
        serializer_class = self.get_serializer_class(serializer_key)
        return Response(serializer_class(instance).data, status=status_code)

    def get_list_response(self, queryset=None, **params):

        """
        Get post response from data.

        Automatically serialize data from action name or parameter serializer_key.

        if data is empty build response from

        Args:
            queryset(QuerySet): data to return in response
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_201_CREATED
        Returns:
            Response
        """
        default = {
            "serializer_key": None,
            "status_code": status.HTTP_200_OK,
        }
        params = {**default, **params}

        serializer_class = self.get_serializer_class(params.get("serializer_key"))
        if queryset is None:
            queryset = self.service.list()

        optimized_queryset = serializer_class.get_optimized_queryset(queryset)
        if self.request.query_params:
            optimized_queryset = optimized_queryset.filter(
                **self.request.query_params.dict()
            )
        serializer = serializer_class(optimized_queryset, many=True)
        return Response(serializer.data, params.get("status_code"))

    def get_retrieve_response(self, instance=None, **params) -> Response:
        """
        Build retrieve response
        Args:
            instance(Model|None): if None use service.get_by_pk
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_201_CREATED

        Returns:
            Response
        """
        return self.get_detail_response(instance=instance, **params)

    def get_create_response(self, instance, **params) -> Response:
        """
        Build create response

        Args:
            instance(Model): model instance created
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_201_CREATED

        Returns:
            Response
        """
        default = {"status_code": status.HTTP_201_CREATED}
        params = {**default, **params}
        return self.get_detail_response(instance=instance, **params)

    def get_update_response(self, instance, **params) -> Response:
        """
        Build update response
        Args:
            instance(Model): model instance updated
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_200_OK

        Returns:
            Response
        """
        return self.get_detail_response(instance=instance, **params)

    def get_delete_response(self, instance=None, **params) -> Response:
        """
        Build delete response. If instance is None => Response status=204

        Args:
            instance(Model|None): model instance deleted
            **params:
                - serializer_key: serializer to use (default: action name)
                - status_code: default is status.HTTP_200_OK ou HTTP_204_NO_CONTENT si
                                instance is None

        Returns:
            Response
        """
        if instance:
            return self.get_detail_response(instance=instance, **params)
        return Response(status=status.HTTP_204_NO_CONTENT)



import pytest
from django.conf import settings

from w.services.technical.models.yousign import YouSignRequest
from w.services.technical.yousign_service import YouSignService
from w.tests.helpers import request_test_helper
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestYouSignService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        data = {
            "procedure_name": "Test de création",
            "procedure_desc": "Un test pour valider le fonctionnement",
            "filename": cls.get_datasets_dir("pdf/exemple_pdf.pdf"),
            "member_firstname": "Jean",
            "member_lastname": "Bono",
            "member_email": "jb@gmail.fr",
            "member_phone": "+33612345678",
            "file_signature_position": "123,456,789,112",
        }
        cls.yousign_request = YouSignRequest(**data)

    @staticmethod
    def setup_method():
        # clear service on each test
        YouSignService.clear()
        YouSignService.init(settings.YOUSIGN_API_URL, settings.YOUSIGN_API_KEY)

    """
    create
    """

    def test_create_with_service_not_initialized_raise_runtime_error(self):
        """ Ensure method raisef RuntimeError if service is not initialized """
        YouSignService.clear()
        match = "Service YouSignService must be initialized first"
        with pytest.raises(RuntimeError, match=match):
            YouSignService.create(self.yousign_request)

    def test_create_with_file_not_found_raise_runtime_error(self):
        """ Ensure method raise RuntimeError if file does not exists """
        # copy valid yousign request
        wrong_request = YouSignRequest(**self.yousign_request.to_dict())
        wrong_request.filename = "unknown.pdf"
        match = f"{wrong_request.filename} does not exists"
        with pytest.raises(RuntimeError, match=match):
            YouSignService.create(wrong_request)

    def test_create_with_failed_upload_raise_runtime_error(self):
        """ Ensure method raise """
        response = {
            "json_file": self.get_datasets_dir("yousign/files_bad_request.json"),
        }
        match = (
            "Failed to upload .*exemple_pdf.pdf.*\(400 - Bad Request\) "
            ': content: Format not allowed for this field. "application\/vnd.'
            'openxmlformats-officedocument.spreadsheetml.sheet" given.'
        )
        with request_test_helper.request_failure(response, method="post"):
            with pytest.raises(RuntimeError, match=match):
                YouSignService.create(self.yousign_request)

    def test_create_with_failed_create_procedure_raise_runtime_error(self):
        """ Ensure method raise RuntimeError """
        responses = request_test_helper.get_response(
            json_file=self.get_datasets_dir("yousign/files_success.json")
        ) + request_test_helper.get_400_response(
            json_file=self.get_datasets_dir("yousign/procedures_bad_request.json")
        )

        match = "Failed to create procedure .*"
        with request_test_helper.mock_request(responses, method="post"):
            with pytest.raises(RuntimeError, match=match):
                YouSignService.create(self.yousign_request)

    def test_create_with_success_return_dict(self):
        """ Ensure method succeed """
        response = {
            "json_file": [
                self.get_datasets_dir("yousign/files_success.json"),
                self.get_datasets_dir("yousign/procedures_success.json"),
            ]
        }
        with request_test_helper.request_success(response, method="post") as m:
            result = YouSignService.create(self.yousign_request)
        self.assert_equals_resultset(
            {"result": result, "mocks": self.get_mock_calls(m)}
        )



from rest_framework import serializers

from w.drf.serializers.mixins import DrfValidatorSerializerMixin
from w.tests.fixtures.datasets.django_app import models


class SimpleCreateValidation(DrfValidatorSerializerMixin, serializers.Serializer):
    integer = serializers.IntegerField()
    string = serializers.CharField()
    date = serializers.DateField()


class ExampleCreateValidation(DrfValidatorSerializerMixin, serializers.ModelSerializer):
    class Meta:
        model = models.ExampleWithInternalDependencies
        fields = "__all__"


class ExampleUpdateValidation(DrfValidatorSerializerMixin, serializers.ModelSerializer):
    class Meta:
        model = models.ExampleWithInternalDependencies
        fields = "__all__"



import factory

from w.tests.fixtures.datasets.builders.models import (
    RelatedDependency,
    ExampleWithAutoNow,
    FakeDjangoModel,
    FakeDjangoModelWithForeignKey,
)
from w.tests.fixtures.datasets.django_app import models as django_models


class SimpleExampleFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = django_models.Example

    attribute_one = "value1"
    attribute_two = "value2"


class InternalDependencyOneFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = django_models.InternalDependencyOne

    name = factory.sequence(lambda n: f"internal_one_name_{n}")


class InternalDependencyTwoFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = django_models.InternalDependencyTwo

    name = factory.sequence(lambda n: f"internal_two_name_{n}")


class ExampleWithInternalDependenciesFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = django_models.ExampleWithInternalDependencies

    name = factory.sequence(lambda n: f"example_internal_name_{n}")
    internal_one = factory.SubFactory(InternalDependencyOneFactory)
    internal_two = factory.SubFactory(InternalDependencyTwoFactory)


class RelatedDependencyFactory(factory.Factory):
    class Meta:
        model = RelatedDependency

    name = factory.sequence(lambda n: f"related_name_{n}")
    example = factory.SubFactory(InternalDependencyOneFactory)


class ExampleWithAutoNowFactory(factory.Factory):
    class Meta:
        model = ExampleWithAutoNow

    attribute_one = "value1"


class FakeDjangoModelFactory(factory.Factory):
    class Meta:
        model = FakeDjangoModel

    id = factory.Sequence(lambda n: n)
    attribute_one = factory.sequence(lambda n: f"attribute_one_{n}")


class FakeDjangoModelWithForeignKeyFactory(factory.Factory):
    class Meta:
        model = FakeDjangoModelWithForeignKey

    id = factory.Sequence(lambda n: n)
    attribute_one = factory.sequence(lambda n: f"name_{n}")
    fk_one = factory.SubFactory(FakeDjangoModelFactory)



from dj_rest_auth.views import LoginView
from django.conf import settings
from rest_framework import status
from rest_framework.response import Response
from rest_framework_api_key.permissions import HasAPIKey
from w.drf.serializers.serpy_serializers import UserSerializer
from w.services.technical.date_service import DateService
from w.utils import import_path


class LoginViewset(LoginView):
    permission_classes = [HasAPIKey]
    _user_serializer = None

    def get_user_serializer(self):
        if not self._user_serializer:
            klass = None
            serializers = getattr(settings, "REST_AUTH_SERIALIZERS", None)
            if serializers and "W_LOGIN_RESPONSE_SERIALIZER" in serializers:
                klass = serializers["W_LOGIN_RESPONSE_SERIALIZER"]
            self._user_serializer = import_path(klass) if klass else UserSerializer
        return self._user_serializer

    def get_response(self):

        if not getattr(settings, "REST_USE_JWT", False):
            raise RuntimeError("This viewset need REST_USE_JWT=True")

        from rest_framework_simplejwt.settings import api_settings as jwt_settings

        jwt_expiration = DateService.to_datetime() + jwt_settings.ACCESS_TOKEN_LIFETIME
        user_serializer = self.get_user_serializer()
        data = {
            "user": user_serializer(self.user).data,
            "access_token": str(self.access_token),
            "expires": DateService.to_mysql_datetime(jwt_expiration),
        }
        response = Response(data, status=status.HTTP_200_OK)
        cookie_name = getattr(settings, "JWT_AUTH_COOKIE", None)

        if cookie_name:
            cookie_secure = getattr(settings, "JWT_AUTH_SECURE", False)
            cookie_httponly = getattr(settings, "JWT_AUTH_HTTPONLY", True)
            cookie_samesite = getattr(settings, "JWT_AUTH_SAMESITE", "Lax")
            jwt_expiration = (
                DateService.to_datetime() + jwt_settings.REFRESH_TOKEN_LIFETIME
            )
            response.set_cookie(
                cookie_name,
                self.refresh_token,
                expires=jwt_expiration,
                secure=cookie_secure,
                httponly=cookie_httponly,
                samesite=cookie_samesite,
            )
        return response



from abc import ABC
from typing import Dict

from django.db import models
from w.services.technical.date_service import DateService
from w.tests.helpers import date_test_helper
from w.tests.mixins.factory_boy_mixin import FactoryBoyMixin


class AbstractTestBuilder(FactoryBoyMixin, ABC):
    _factories: Dict = {}
    _main_factory_key = None
    _related_built: Dict = {}

    @classmethod
    def _check_factory_exists(cls, factory_key):
        if factory_key not in cls._factories:
            raise RuntimeError(
                f"{factory_key} missing, add it to {__class__}._factories"
            )

    @classmethod
    def get_factory(cls, factory_key):
        cls._check_factory_exists(factory_key)
        return cls._factories[factory_key]

    def __init__(self):
        # attributes to use with main factory
        self._attrs = {}
        # attributes to build for foreign key factories
        self._fk_factories_attrs = {}
        # builder for foreign key
        self._fk_builders = {}
        # related factories and its attributes
        self._related_factories_attrs = {}
        # define today date
        self._date_today = None

    def _complete_attrs(self, attrs):
        """
        For each registered factories, build attribute from factory and add it to
        attrs

        Args:
            attrs:

        Returns:
            dict
        """
        attrs = {**self._attrs, **attrs}

        for attr, params in self._fk_factories_attrs.items():
            attrs[attr] = self.get_factory(params["key"])(**params["attrs"])

        for attr, builder in self._fk_builders.items():
            attrs[attr] = builder.build()

        return attrs

    def _set_foreign_key_factory_attrs(
        self, attr: str, factory_attrs: dict, factory_key=None
    ):
        """
        Define foreign key factory
        Args:
            attr(str): foreign key attribute
            factory_attrs(dict): attributes for factory
            factory_key(str|None): if None factory_key = attr
        """
        if factory_key is None:
            factory_key = attr
        self._fk_factories_attrs[attr] = {"key": factory_key, "attrs": factory_attrs}

    def _set_foreign_key_builder(self, attr: str, builder):
        self._fk_builders[attr] = builder

    def _create_related_factory_if_missing(self, attr: str, factory_key=None):
        if factory_key is None:
            factory_key = attr
        if attr not in self._related_factories_attrs:
            self._related_factories_attrs[attr] = {"key": factory_key, "list_attrs": []}

    def _add_related_factory_attrs(
        self, attr: str, factory_attrs: dict, factory_key=None
    ):
        self._create_related_factory_if_missing(attr, factory_key)
        self._related_factories_attrs[attr]["list_attrs"].append(factory_attrs)

    def _set_related_factory_attrs(
        self, attr: str, list_factory_attrs: list, factory_key=None
    ):
        self._create_related_factory_if_missing(attr, factory_key)
        self._related_factories_attrs[attr]["list_attrs"] += list_factory_attrs

    def _set_nb_related_factory_attrs(
        self, attr: str, nb, factory_attrs: dict, factory_key=None
    ):
        self._create_related_factory_if_missing(attr, factory_key)
        for _ in range(nb):
            self._related_factories_attrs[attr]["list_attrs"].append(factory_attrs)

    def _build_related_factory(self, factory_key, factory_attrs, related_attrs):
        factory_attrs = {**factory_attrs, **related_attrs}
        return self.get_factory(factory_key)(**factory_attrs)

    def _build_related_factories(self, related_attrs: dict) -> dict:
        """
        Build related factories linked on linked_attrs
        Args:
            related_attrs(dict): {<related_attr>: value, ...}

        Returns:
            dict: list built models by factory_key
        """
        factory_builds = {}
        for attr, params in self._related_factories_attrs.items():
            factory_builds[attr] = []
            for factory_attrs in params["list_attrs"]:
                factory_builds[attr].append(
                    self._build_related_factory(
                        params["key"], factory_attrs, related_attrs
                    )
                )
        return factory_builds

    def _get_related_built(self, factory_key: str) -> list:
        return self._related_built[factory_key]

    @classmethod
    def list_used_builders(cls):
        """ List builders used by this builder for test reset sequence """
        return [cls]

    def with_today_is(self, date):
        """
        Force today to be date

        Args:
            date (str|Arrow|datetime.datetime|datetime.date): date today

        Returns:
            self
        """
        self._date_today = DateService.to_datetime(date)
        return self

    def with_attrs(self, **attrs):
        """
        Define main attributes
        """
        self._attrs = {**self._attrs, **attrs}
        return self

    def _build(self, attrs):
        main_built = self.get_factory(self._main_factory_key)(**attrs)
        self._related_built = self._build_related_factories(
            {self._main_factory_key: main_built}
        )
        return main_built

    def build(self, **attrs):
        """
        Build main model
        """
        if self._main_factory_key is None:
            raise RuntimeError("You must define '_main_factory_key' in your builder")

        attrs = self._complete_attrs(attrs)

        if self._date_today:
            with date_test_helper.today_is(self._date_today):
                return self._build(attrs)

        return self._build(attrs)

    def build_multiple(self, nb, **attrs):
        """ Build multiple """
        builds = []
        for _ in range(0, nb):
            builds.append(self.build(**attrs))
        return builds

    @staticmethod
    def _is_django_model(value) -> bool:
        """ Check if value is django model """
        return isinstance(value, models.Model)

    def build_create_data(self, **attrs) -> dict:
        """ Build create/update data from main factory """
        import factory

        attrs = self._complete_attrs(attrs)
        factory_class = self.get_factory(self._main_factory_key)

        # looking for SubFactory to build them
        list_attrs = [
            a
            for a in factory_class.__dict__.keys()
            if not a.startswith("_") and a not in attrs
        ]
        for attr in list_attrs:
            value = getattr(factory_class, attr)
            if isinstance(value, factory.SubFactory):
                attrs[attr] = value.factory_wrapper.factory()

        return factory.build(dict, FACTORY_CLASS=factory_class, **attrs)

    def build_post_data(self, **attrs) -> dict:
        """ Build post data from main factory """
        if self._date_today is None:
            built = self.build_create_data(**attrs)
        else:
            with date_test_helper.today_is(self._date_today):
                built = self.build_create_data(**attrs)

        result = {}
        for attr, value in built.items():
            # remove foreign key object
            if self._is_django_model(value):
                result[f"{attr}"] = value.pk
            else:
                result[attr] = value

        return result



import pytest
from openpyxl import Workbook

from w.services.technical.excel_service import ExcelService
from w.tests.mixins.testcase_mixin import TestCaseMixin
from w.exceptions import ValidationError


class TestExcelService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.filename = cls.get_datasets_dir("excel/Excel2010.xlsx")
        cls.mapping = {"header1": "columnA", "header2": "columnB", "header3": "columnC"}

    """
    is_excel
    """

    def test_is_excel_with_no_excel_file_return_false(self):
        """ Ensure method return False on no excel file """
        files = ["toto.txt", "excel.txt", "excel.xls.gz"]
        for file in files:
            assert ExcelService.is_excel(file) is False

    def test_is_excel_with_success_return_true(self):
        """ Ensure method return True on excel file """
        files = ["toto.xls", "excel.xlsx"]
        for file in files:
            assert ExcelService.is_excel(file) is True

    """
    open_workbook
    """

    def test_open_workbook_with_file_not_exists_raise_runtime_error(self):
        """ Ensure RuntimeError is raised if filename does not exists """
        with pytest.raises(RuntimeError, match="unknown-file does not exists"):
            ExcelService.open_workbook("unknown-file")

    def test_open_workbook_with_invalid_excel_suffix_raise_runtime_error(self):
        """ Ensure method raise runtime error if file is not excel """
        filename = self.get_datasets_dir("csv/simple.csv")
        match = "simple.csv is not a valid excel file"
        with pytest.raises(RuntimeError, match=match):
            ExcelService.open_workbook(filename)

    def test_open_workbook_with_not_valid_excel_raise_runtime_error(self):
        """ Ensure RuntimeError is raised if filename is not valid excel """
        filename = self.get_datasets_dir("excel/invalid_excel.xls")
        match = "invalid_excel.xls is not a valid excel file"
        with pytest.raises(RuntimeError, match=match):
            ExcelService.open_workbook(filename)

    def test_open_workbook_with_success_return_book(self):
        """ Ensure method succeed return Book """
        actual = ExcelService.open_workbook(self.filename)
        assert isinstance(actual, Workbook)

    """
    load
    """

    def test_load_with_invalid_excel_suffix_raise_runtime_error(self):
        """ Ensure method raise runtime error if file is not excel """
        filename = self.get_datasets_dir("csv/simple.csv")
        match = "simple.csv is not a valid excel file"
        with pytest.raises(RuntimeError, match=match):
            ExcelService.load(filename, self.mapping)

    def test_load_with_not_valid_excel_raise_runtime_error(self):
        """ Ensure RuntimeError is raised if filename is not valid excel """
        filename = self.get_datasets_dir("excel/invalid_excel.xls")
        match = "invalid_excel.xls is not a valid excel file"
        with pytest.raises(RuntimeError, match=match):
            ExcelService.load(filename, self.mapping)

    def test_load_with_invalid_header_raise_validation_error(self):
        """ Ensure method raise exceptions validation error if header is false """
        mapping = self.mapping.copy()
        mapping["colWrong"] = mapping.pop("header3")
        match = "Unprocessable Entity"
        with pytest.raises(ValidationError, match=match):
            ExcelService.load(self.filename, mapping)

    def test_load_with_success_return_list(self):
        """ Ensure method succeed """
        self.assert_equals_resultset(ExcelService.load(self.filename, self.mapping))

    def test_load_with_empty_col_success_return_list(self):
        """ Ensure method succeed even with an empty column """
        self.assert_equals_resultset(
            ExcelService.load(
                self.get_datasets_dir("excel/Excel2010_with_empty_col.xlsx"),
                self.mapping,
            )
        )

    def test_load_with_specific_bad_sheet_name_raise_runtime_error(self):
        """ Ensure method raise runtime error if sheet name is bad """
        match = "bad_sheet_name is not a valid sheet name"

        with pytest.raises(RuntimeError, match=match):
            ExcelService.load(self.filename, self.mapping, "bad_sheet_name")

    def test_load_with_specific_sheet_name_success_return_list(self):
        """ Ensure method succeed with specific sheet name """
        mapping = {
            "header1_sheet2": "columnA",
            "header2_sheet2": "columnB",
            "header3_sheet2_with_space": "columnC",
        }

        self.assert_equals_resultset(
            ExcelService.load(self.filename, mapping, "Sheet_2")
        )



from rest_framework.exceptions import PermissionDenied


def group_required(groups):
    """
    Limit api access to list of groups. User must be authenticated too.

    If user do not belongs to one of the groups return 403 permission denied.

    Usage:
        @group_required(("group1", "group2", ...))
        def list(self, request, *args, **kwargs):
            ...
    """

    def decorator(drf_custom_method):
        def _decorator(self, *args, **kwargs):
            if self.request.user.groups.filter(name__in=groups).exists():
                return drf_custom_method(self, *args, **kwargs)
            raise PermissionDenied()

        return _decorator

    return decorator



import abc

from rest_framework import status

from w.django.utils import reverse


# noinspection PyUnresolvedReferences
class ApiViewSetMixin:
    resource = None
    _calling_filename = None

    @staticmethod
    @abc.abstractmethod
    def get_resource_objects():
        pass

    @classmethod
    def get_resource_pk(cls):
        return getattr(cls, cls.resource).pk

    @classmethod
    def get_post_data_resource(cls):
        return getattr(cls, "get_rest_data_%s" % cls.resource)(rest_method="POST")

    @classmethod
    def get_put_data_resource(cls):
        return getattr(cls, "get_rest_data_%s" % cls.resource)(rest_method="PUT")

    @classmethod
    def get_patch_data_resource(cls):
        return getattr(cls, "get_rest_data_%s" % cls.resource)(rest_method="PATCH")

    """
    GET /{resources}
    """

    def _list_resource_with_no_registered_user_return_401(self):
        """
        Ensure we cannot get resource if no user authenticated
        """
        self.assert_get_with_no_registered_user_return_401("%s-list" % self.resource)

    def _list_resource_with_success_return_200(self, user=None):
        """
        Ensure we can get all resources
        """
        url = reverse("%s-list" % self.resource)
        response = self.client_get(url, user=user)
        self.assert_get_response(response, calling_filename=self._calling_filename)

    def _list_resource_return_405(self):
        """
        Ensure we cannot list all resources
        """
        url = reverse("%s-list" % self.resource)
        response = self.client_get(url)
        self.assert_rest_status_code(response, status.HTTP_405_METHOD_NOT_ALLOWED)

    """
    GET /{resources}/{pk}
    """

    def _get_resource_with_no_registered_user_return_401(self):
        """
        Ensure we cannot get resource if no user authenticated
        """
        self.assert_get_with_no_registered_user_return_401(
            "%s-detail" % self.resource, {"pk": 1}
        )

    def _get_resource_with_success_return_200(self, user=None):
        """
        Ensure we can get a resource
        """
        url = reverse(
            "%s-detail" % self.resource, kwargs={"pk": self.get_resource_pk()}
        )
        response = self.client_get(url, user=user)
        self.assert_get_response(response, calling_filename=self._calling_filename)

    def _get_resource_with_me_return_200(self, user=None):
        """
        Ensure we can get resource linked to current user
        """
        url = reverse("%s-detail" % self.resource, kwargs={"pk": "me"})
        response = self.client_get(url, user=user)
        self.assert_get_response(response, calling_filename=self._calling_filename)

    """
    POST /{resources}
    """

    def _post_resource_with_no_registered_user_return_401(self):
        """
        Ensure we cannot post resource if no user authenticated
        """
        self.assert_post_with_no_registered_user_return_401("%s-list" % self.resource)

    def _post_resource_with_success_return_201(self, user=None):
        """
        Ensure we can post resource
        """
        url = reverse("%s-list" % self.resource)
        data = self.get_post_data_resource()
        response = self.client_post(url, data=data, user=user)
        self.assert_post_response(
            data, response, calling_filename=self._calling_filename
        )

    """
    PUT /{resources}/{pk}
    """

    def _put_resource_with_no_registered_user_return_401(self):
        """
        Ensure we cannot put resource if no user authenticated
        """
        self.assert_put_with_no_registered_user_return_401(
            "%s-detail" % self.resource, {"pk": 1}
        )

    def _put_resource_with_success_return_200(self, user=None):
        """
        Ensure we can put a resource
        """
        url = reverse(
            "%s-detail" % self.resource, kwargs={"pk": self.get_resource_pk()}
        )
        data = self.get_put_data_resource()
        response = self.client_put(url, data=data, user=user)
        self.assert_put_response(
            self.get_resource_pk(),
            data,
            response,
            calling_filename=self._calling_filename,
        )

    """
    PATCH /{resources}/{pk}
    """

    def _patch_resource_with_no_registered_user_return_401(self):
        """
        Ensure we cannot patch resource if no user authenticated
        """
        self.assert_put_with_no_registered_user_return_401(
            "%s-detail" % self.resource, {"pk": 1}
        )

    def _patch_resource_with_success_return_200(self, user=None):
        """
        Ensure we can patch a resource
        """
        url = reverse(
            "%s-detail" % self.resource, kwargs={"pk": self.get_resource_pk()}
        )
        data = self.get_patch_data_resource()
        response = self.client_patch(url, data=data, user=user)
        self.assert_patch_response(
            self.get_resource_pk(),
            data,
            response,
            calling_filename=self._calling_filename,
        )

    """
    DEL /{resources}/{pk}
    """

    def _del_resource_with_no_registered_user_return_401(self):
        """
        Ensure we cannot delete resource if no user authenticated
        """
        self.assert_delete_with_no_registered_user_return_401(
            "%s-detail" % self.resource, {"pk": 1}
        )

    def _del_issue_with_success_return_204(self, user=None):
        """
        Ensure we can delete an issue
        """
        url = reverse(
            "%s-detail" % self.resource, kwargs={"pk": self.get_resource_pk()}
        )
        response = self.client_delete(url, user=user)
        self.assert_delete_response(self.get_resource_pk(), response)



# flake8: noqa F405
from typing import Type

from django.db.models import Prefetch, QuerySet
from django.utils.translation import gettext as _
from serpy import *  # noqa

from w.services.technical.date_service import DateService


class DateField(Field):
    def to_value(self, value):
        if value:
            return DateService.to_mysql_date(value)


class DatetimeField(Field):
    def to_value(self, value):
        if value:
            return DateService.to_mysql_datetime(value)


class ManyToManyField(Field):
    def __init__(
        self,
        serializer,
        attr=None,
        call=False,
        label=None,
        required=True,
    ):
        super().__init__(attr, call, label, required)
        self.serializer = serializer

    def to_value(self, value):
        if value and value.exists():
            return self.serializer(value.all(), many=True).data
        return []


class PhoneNumberField(Field):
    def to_value(self, value):
        if value:
            return value.as_e164


class FileField(Field):
    def to_value(self, value):
        return value.name if value else None


class TranslateField(Field):
    def to_value(self, value):
        return _(value) if value else ""


class CloudinaryField(Field):
    def to_value(self, value):
        return value.url if value else None


class SerpySerializer(Serializer):
    """
    Data serializer built on Serpy.

    It adds query_string optimization and add custom Fields

    @see serpy documentation
    """

    @classmethod
    def get_optimized_queryset(cls, qs, prefix_related=None) -> QuerySet:
        """
        Get optimal QuerySet for serialization

        Args:
            qs (QuerySet): query set to complete
            prefix_related (str): prefix to add (for internal usage)

        Returns:
            QuerySet
        """

        if cls.select_related():
            qs = qs.select_related(*cls._get_select_related(prefix_related))
        if cls.prefetch_related():
            qs = qs.prefetch_related(*cls._get_prefetch_related(prefix_related))
        for prefix, serializer in cls.serializer_related().items():
            prefix = f"{prefix_related}__{prefix}" if prefix_related else prefix
            qs = serializer.get_optimized_queryset(qs, prefix)
        return qs

    @classmethod
    def select_related(cls) -> list:
        """
        list of related attribute to select
        Returns:
            list
        """
        return []

    @classmethod
    def prefetch_related(cls) -> list:
        """
        list of related of attribute to prefetch
        Returns:
            list
        """
        return []

    @classmethod
    def serializer_related(cls) -> dict:
        """
        List of related serializer
        Returns:
            dict: {"<attribute>": <serializer class>, ...}
        """
        return {}

    @classmethod
    def _get_select_related(cls, prefix=None) -> list:
        """
        get related select prefixed if needed
        Args:
            prefix (str): prefix to add

        Returns:
            list
        """
        if prefix is None:
            return cls.select_related()
        return [f"{prefix}__{r}" for r in cls.select_related()]

    @classmethod
    def _get_prefetch_related(cls, prefix=None):
        """
        get related prefetc prefixed if needed
        Args:
            prefix (str): prefix to add

        Returns:
            list
        """
        if prefix is None:
            return cls.prefetch_related()

        result = []
        for r in cls.prefetch_related():
            if not isinstance(r, Prefetch):
                r = f"{prefix}__{r}"
            else:
                r.prefetch_through = f"{prefix}__{r.prefetch_through}"
                r.prefetch_to = f"{prefix}__{r.prefetch_to}"
            result.append(r)
        return result



from django.db import models
from typing import List
from w.services.abstract_service import AbstractService
from w import exceptions


# noinspection PyCallingNonCallable
from w.services.technical.dict_service import DictService


class AbstractModelService(AbstractService):
    _model: models.Model

    @staticmethod
    def _is_field(field):
        """Check if field from model._meta.get_fields() is field"""
        return hasattr(field, "related_name") is False

    @classmethod
    def list_fields(cls) -> list:
        """List model fields"""
        return [f.name for f in cls._model._meta.get_fields() if cls._is_field(f)]

    @classmethod
    def clean_attrs(cls, attrs):
        """Remove unexpected model attributes"""
        return DictService.keep_keys(attrs, cls.list_fields())

    @classmethod
    def create(cls, **attrs):
        """
        Create model instance

        Args:
            **attrs: model attributes values

        Returns:
            models.Model
        """
        instance = cls._model(**attrs)
        instance.save()
        return instance

    @classmethod
    def get_by_pk(cls, pk):
        """
        Retrieve model by its primary key

        Returns:
            Model
        """
        return cls._model.objects.get(pk=pk)

    @classmethod
    def get_if_exists(cls, **filters):
        """Retrieve instance if exists else return None"""
        qs = cls._model.objects.filter(**filters)
        if qs.exists():
            return qs.first()
        return None

    @classmethod
    def is_exists_by_pk(cls, pk) -> bool:
        """
        Check model existsby its primary key

        Returns
            bool
        """
        qs = cls._model.objects.filter(pk=pk)
        return qs.exists()

    @classmethod
    def check_by_pk(cls, pk):
        """
        Check model exists by its primary key

        if found return model else raise NotFoundError

        Raises
            NotFoundError
        """
        try:
            return cls._model.objects.get(pk=pk)
        except cls._model.DoesNotExist:
            label = cls._model._meta.verbose_name.title()  # noqa
            raise exceptions.NotFoundError(f"{label} not found (pk={pk})")

    @classmethod
    def list(cls, **filters) -> models.QuerySet:
        """
        List models filtered by filters (optional)

        Args:
            **filters: filter result

        Returns:
            QuerySet
        """
        if filters:
            return cls._model.objects.filter(**filters)
        return cls._model.objects.all()

    @classmethod
    def list_pks(cls, **filters) -> List:
        """
        List model pks filtered by filters (optional)

        Args:
            **filters:

        Returns:
            List
        """
        return cls.list(**filters).values_list("pk", flat=True)

    @classmethod
    def update(cls, instance, **attrs):
        """
        Update model instance

        Args:
            instance: model instance to update
            **attrs: model attributes values

        Returns:
            models.Model
        """
        update_it = False
        for attr, value in attrs.items():
            if getattr(instance, attr) != value:
                setattr(instance, attr, value)
                update_it = True

        if update_it:
            instance.save()
        return instance

    @classmethod
    def delete(cls, filters) -> int:
        """
        Delete

        Args:
            filters: model filters to delete

        Returns:
            nb deleted
        """
        nb, _ = cls._model.objects.filter(**filters).delete()
        return nb

    @classmethod
    def delete_by_pk(cls, pk) -> int:
        """
        Delete a model by is primary key

        Args:
            pk: model pk to delete

        Returns:
            nb deleted
        """
        return cls.delete({"pk": pk})

    @classmethod
    def to_dict(cls, instance):
        return {f: getattr(instance, f) for f in cls.list_fields()}



import factory


class FakerMixin:
    @classmethod
    def faker(cls, provider, **kwargs):
        """
        Wrapper for 'faker' values.

        Args:
            provider(str): the name of the Faker field
            **kwargs: provider params
        """
        return factory.Faker(provider, **kwargs).generate()

    @classmethod
    def fake_str(cls, length) -> str:
        """ Generate random str of length characters """
        return cls.faker("pystr", min_chars=length, max_chars=length)



from django.db import models

from w.django.models import AbstractCreatedUpdatedModel


class Author(models.Model):
    firstname = models.CharField(max_length=50)
    lastname = models.CharField(max_length=50)


class Series(models.Model):
    name = models.CharField(max_length=128)


class Book(models.Model):
    name = models.CharField(max_length=128)
    author = models.ForeignKey(Author, on_delete=models.CASCADE, related_name="books")
    series = models.ForeignKey(
        Series, on_delete=models.CASCADE, related_name="books", blank=True, null=True
    )


class Character(models.Model):
    firstname = models.CharField(max_length=50)
    lastname = models.CharField(max_length=50)
    books = models.ManyToManyField(Book, related_name="characters")


class AutoNowModel(AbstractCreatedUpdatedModel):
    pass



import base64
import json

from django.conf import settings

from w.services.abstract_service import AbstractService
from w.services.technical.filesystem_service import FilesystemService
from w.services.technical.json_service import JsonService
from w.services.technical.models.yousign import YouSignRequest
from w.services.technical.request_service import RequestService


class YouSignService(AbstractService):
    _api_url = None
    _request_session = None
    _headers = None

    @classmethod
    def init(cls, api_url=None, api_key=None) -> None:
        """ Initialize service with API url and key """
        if api_url is None:
            api_url = settings.YOUSIGN_API_URL
        if api_key is None:
            api_key = settings.YOUSIGN_API_KEY

        cls._api_url = api_url
        cls._headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }

    @classmethod
    def clear(cls):
        """ Clear service """
        super().clear()
        cls._api_url = None
        cls._request_session = None

    @classmethod
    def _is_initialized(cls) -> bool:
        """ Check service is initialized """
        return cls._api_url and cls._headers

    @classmethod
    def _raise_api_failed(cls, response, msg):
        """
        raise RuntimeError for yousign api failure
        """
        if JsonService.is_valid(response.orig_content):
            error = json.loads(response.orig_content)
            error = error["detail"]
        else:
            error = "no detail"

        raise RuntimeError(
            f"{msg} ({response.status_code} - {response.content}) : {error}"
        )

    @classmethod
    def _upload_file(cls, filename):
        """
        Upload file to YouSign (first step)

        Args:
            filename(str): file to upload

        Returns:
            file_id: file id
        """
        payload = {"name": FilesystemService.get_basename(filename), "content": None}
        with open(filename, "rb") as f:
            payload["content"] = base64.b64encode(f.read())
        payload["content"] = payload["content"].decode("utf-8")
        payload = JsonService.dump(payload)
        response = RequestService.post(
            "/files", data=payload, session=cls._request_session
        )
        if response.success:
            return response.content["id"]

        cls._raise_api_failed(response, f"Failed to upload '{filename}'")

    @classmethod
    def _create_procedure(cls, file_id, data: YouSignRequest) -> str:
        """
        Create procedure

        Args:
            file_id(str): uploaded file id for procedure
            data(YouSignRequest): needed data to create procedure

        Returns:
            str: member id
        """
        payload = data.get_procedure_payload(file_id)
        payload = JsonService.dump(payload)
        response = RequestService.post(
            "/procedures", data=payload, session=cls._request_session
        )
        if response.success:
            return response.content["members"][0]["id"]
        cls._raise_api_failed(response, "Failed to create procedure")

    @classmethod
    def create(cls, data: YouSignRequest):
        """
        create signature procedure

        Args:
            data(YouSignRequest): data needed to start procedure

        Returns:
            str: member id needed for iFrame
        """
        cls._check_is_initialized()
        FilesystemService.check_file_exists(data.filename)
        cls._request_session = RequestService.init_session(
            cls._api_url, headers=cls._headers
        )

        # basic mode
        # Step 1 - Upload the file
        file_id = cls._upload_file(data.filename)

        # Step 2 - Create the procedure
        return cls._create_procedure(file_id, data)



from w.services.abstract_service import AbstractService
import json
import time
import urllib.error
import urllib.parse
import urllib.request
from django.conf import settings

API_KEY = settings.GOOGLE_MAP_SECRET
GEOCODE_BASE_URL = "https://maps.googleapis.com/maps/api/geocode/json"
AUTOCOMPLETE_BASE_URL = (
    "https://maps.googleapis.com/maps/api/place/queryautocomplete/json"
)
PLACE_DETAIL_URL = "https://maps.googleapis.com/maps/api/place/details/json"
"""
"""


class GoogleMapService(AbstractService):
    @classmethod
    def geocode_address(cls, textual_address):
        if settings.MOCK_GMAP_CALLS:
            return {"lat": 10, "lng": 10}

        url = cls.build_geocode_url(textual_address)

        return cls._loop_until_endpoint_is_reached(url, cls._parse_geocode_response)

    @classmethod
    def autocomplete_address(cls, textual_address, session_token=None):
        if settings.MOCK_GMAP_CALLS:
            return {"result": "1 rue du test", "location": {"lat": 10, "lng": 10}}

        url = cls.build_autocomplete_place_url(textual_address, session_token)

        return cls._loop_until_endpoint_is_reached(
            url, cls._parse_autocomplete_place_response
        )

    @classmethod
    def get_place_detail(cls, place_id):
        url = cls.build_place_detail_url(place_id)

        return cls._loop_until_endpoint_is_reached(
            url, cls._parse_place_detail_response
        )

    @classmethod
    def build_place_detail_url(cls, place_id):
        # Join the parts of the URL together into one string.
        return (
            f"{PLACE_DETAIL_URL}?place_id={place_id}&key="
            f"{API_KEY}&fields=geometry,formatted_address,place_id"
        )

    @classmethod
    def build_autocomplete_place_url(cls, address, session_token=None):
        # Join the parts of the URL together into one string.
        urls_params = {
            "input": f"{address}",
            "key": API_KEY,
            "langage": "fr",
            "components": "country:fr|country:ch|country:be|country:de|country:es",
        }
        if session_token:
            urls_params.update(sessiontoken=session_token)
        params = urllib.parse.urlencode(urls_params)
        return f"{AUTOCOMPLETE_BASE_URL}?{params}"

    @classmethod
    def build_geocode_url(cls, address):
        # Join the parts of the URL together into one string.
        params = urllib.parse.urlencode({"address": f"{address}", "key": API_KEY})
        return f"{GEOCODE_BASE_URL}?{params}"

    @classmethod
    def _loop_until_endpoint_is_reached(cls, url, parse_result_func):
        current_delay = 0.1  # Set the initial retry delay to 100ms.
        max_delay = 5  # Set the maximum retry delay to 5 seconds.
        while True:
            try:
                # Get the API response.
                response = urllib.request.urlopen(url)
            except urllib.error.URLError:
                pass  # Fall through to the retry loop.
            else:
                # If we didn't get an IOError then parse the result.
                return parse_result_func(response)

            if current_delay > max_delay:
                raise RuntimeError("Too many retry attempts.")
                # print("Waiting", current_delay, "seconds before retrying.")
                time.sleep(current_delay)
                current_delay *= 2  # Increase the delay each time we retry.

    @classmethod
    def _parse_geocode_response(cls, response_to_parse):
        result = json.load(response_to_parse)
        if result["status"] == "OK":
            return result["results"][0]["geometry"]["location"]
        elif result["status"] == "ZERO_RESULTS":
            return None
        elif result["status"] != "UNKNOWN_ERROR":
            # Many API errors cannot be fixed by a retry, e.g. INVALID_REQUEST
            raise RuntimeError(result["error_message"])

    @classmethod
    def _parse_autocomplete_place_response(cls, response_to_parse):
        result = json.load(response_to_parse)
        results_with_id = [
            p for p in result["predictions"] if p.get("place_id", None) is not None
        ]
        if result["status"] == "OK":
            parsed_results = list()
            for r in results_with_id:
                if r.get("structured_formatting", None) is None:
                    continue

                if r["structured_formatting"].get("main_text", None) is None:
                    continue

                if r["structured_formatting"].get("secondary_text", None) is not None:
                    parsed_results.append(
                        {
                            "main_text": r["structured_formatting"]["main_text"],
                            "secondary_text": r["structured_formatting"][
                                "secondary_text"
                            ],
                            "place_id": r["place_id"],
                        }
                    )
                else:
                    parsed_results.append(
                        {
                            "main_text": r["structured_formatting"]["main_text"],
                            "secondary_text": "",
                            "place_id": r["place_id"],
                        }
                    )
            return parsed_results
        elif result["status"] == "ZERO_RESULTS":
            return None
        elif result["status"] != "UNKNOWN_ERROR":
            # Many API errors cannot be fixed by a retry, e.g. INVALID_REQUEST
            raise RuntimeError(result["error_message"])

    @classmethod
    def _parse_place_detail_response(cls, response_to_parse):
        result = json.load(response_to_parse)
        if result["status"] == "OK":
            return result["result"]
        elif result["status"] == "ZERO_RESULTS":
            return None
        elif result["status"] != "UNKNOWN_ERROR":
            # Many API errors cannot be fixed by a retry, e.g. INVALID_REQUEST
            raise RuntimeError(result["error_message"])



from abc import abstractmethod

from django.conf import settings
from w.services.abstract_service import AbstractService
from w.services.imports.models.import_report import ImportReport, ImportRowError
from w.services.technical.csv_service import CsvService
from w.services.technical.excel_service import ExcelService
from w.services.technical.filesystem_service import FilesystemService


class AbstractImportService(AbstractService):
    sheet_name = None
    _filename = None
    _context = None

    @classmethod
    def import_file(cls, filename, context=None) -> ImportReport:
        """
        Import filename

        Args:
            filename(str): csv or excel file
            context(dict): context to use for subsequent methods

        Returns:
            Tuple[list, ImportReport]
        """
        FilesystemService.check_file_exists(filename)
        cls._filename = filename
        cls._context = context
        data = cls._load_data(filename, cls._list_mapping_columns())
        return cls.import_data(data)

    @classmethod
    def import_data(cls, data) -> ImportReport:
        """
        Import data

        Args:
            data(dict): data to import

        Returns:
            Tuple[list, ImportReport]: validation_errors, import_report
        """
        validated_datas, validation_errors = cls._validate(data)
        cls._import(validated_datas)
        return cls._get_import_report(validation_errors)

    @classmethod
    def _get_import_report(cls, validation_errors) -> ImportReport:
        """
        Get import report

        Args:
            validation_errors:

        Returns:
            ImportReport
        """
        report = ImportReport(nb_total=len(validation_errors))
        mapping = cls._list_mapping_attributes()
        for idx, error in enumerate(validation_errors):
            if error:
                report.nb_ignored += 1
                report.nb_errors += len(error.keys())
                report.errors.append(
                    ImportRowError(
                        num_row=idx + 1,
                        errors=[
                            f"{mapping[attr]}: {msg[0]}" for attr, msg in error.items()
                        ],
                    )
                )
            else:
                report.nb_imported += 1
        return report

    @staticmethod
    def get_example_dir(relative=True):
        """ get import example directory """
        example_dir = "core/media/imports/examples"
        if relative:
            return example_dir
        return f"{settings.ROOT_DIR}/{example_dir}"

    @classmethod
    def _load_data(cls, filename, mapping_columns):
        """
        Load data from csv or excel file
        Args:
            filename(str): csv or excel file

        Returns:
            list

        Raises:
            RuntimeError:
        """
        filename = FilesystemService.get_path(filename)
        if CsvService.is_csv(filename):
            return CsvService.load(filename, mapping_columns)
        if ExcelService.is_excel(filename):
            return ExcelService.load(filename, mapping_columns, cls.sheet_name)
        raise RuntimeError(f"Format {filename.name} not supported")

    @classmethod
    def _list_mapping_attributes(cls) -> dict:
        """
        List mapping from attributes to csv/excel columns

        Returns:
            dict: {<import header label>: <wanted attribut>, ...}
        """
        return {v: k for k, v in cls._list_mapping_columns().items()}

    @classmethod
    @abstractmethod
    def _list_mapping_columns(cls) -> dict:
        """
        List mapping columns from csv/excel to wanted attributes

        Returns:
            dict: {<import header label>: <wanted attribut>, ...}
        """
        pass

    @classmethod
    @abstractmethod
    def _validate(cls, data) -> tuple:
        """
        Validate data

        Args:
            data(dict): data to validate

        Returns:
            tuple: validated_data, validation_errors
        """
        pass

    @classmethod
    @abstractmethod
    def _import(cls, data) -> None:
        """
        Import data

        Args:
            data([dict]): datas to  import

        Returns:
            None
        """
        pass



from contextlib import contextmanager
from pathlib import Path
from unittest.mock import Mock

from w.services.technical.models.sso import SsoValidToken
from w.services.technical.sso_service import SsoService
from w.tests.helpers import service_test_helper
from w.tests.mixins.testcase_mixin import TestCaseMixin


def _get_dataset(relative_dataset):
    current_dir = Path(__file__).parent
    dataset_filename = current_dir.joinpath(
        "../fixtures/datasets", relative_dataset
    ).resolve()
    return TestCaseMixin._get_dataset(dataset_filename.name, dataset_filename)


# noinspection PyProtectedMember
def _sso_instrospect_mock_attrs(dataset="introspect_success"):
    return {
        "service": SsoService,
        "method_name": "_sso_introspect",
        "return_value": _get_dataset(f"sso/{dataset}.json"),
    }


def _mock_get_or_create_user():
    return {
        "service": SsoService,
        "method_name": "get_or_create_user",
        "return_value": _get_dataset(
            "sso_service/create_sso_user_with_success_return_dict.json"
        ),
    }


@contextmanager
def mock_keycloak_admin_init():
    keycloak_admin = Mock()
    mock_keycloak_admin = {
        "service": SsoService,
        "method_name": "_get_keycloak_admin",
        "return_value": keycloak_admin,
    }
    with service_test_helper.mock_service(**mock_keycloak_admin):
        yield keycloak_admin


@contextmanager
def valid_token_failure():
    mock_attrs = _sso_instrospect_mock_attrs("introspect_failure")
    with service_test_helper.mock_service(**mock_attrs) as m:
        yield m


@contextmanager
def sso_introspect_success():
    mock_attrs = _sso_instrospect_mock_attrs()
    with service_test_helper.mock_service(**mock_attrs) as m:
        yield m


@contextmanager
def valid_token_success(uuid="fake-uuid"):
    fake_decoded_token = {
        "sub": uuid,
        "given_name": "fake-sub",
        "family_name": "fake-given_name",
        "username": "fake-family_name@fake-mail.com",
        "email": "fake-username@fake-mail.com",
        "resource_access": {
            "account": {
                "roles": ["manage-account", "manage-account-links", "view-profile"]
            }
        },
    }
    mock_attrs = {
        "service": SsoService,
        "method_name": "is_token_valid",
        "return_value": SsoValidToken(fake_decoded_token),
    }
    with service_test_helper.mock_service(**mock_attrs) as m:
        yield m


@contextmanager
def get_or_create_user_success():
    with service_test_helper.mock_service(**_mock_get_or_create_user()) as m:
        yield m



"""
Add it to to your Django settings

REST_FRAMEWORK = {
  ...
  'DEFAULT_AUTHENTICATION_CLASSES': [
    ...
    'rest_framework.authentication.SessionAuthentication',
    'w.drf.sso.sso_auth_backend.SsoAuthBackend',
  ]
}

Need extra django settings configuration

SSO={
    "SERVER_URL": <sso url>,
    "REALM": <sso api realm>,
    "CLIENT_ID": <sso client id api>,
    "CLIENT_SECRET_KEY": <sso client secret see Client credentials>,
    "USER_SERVICE": <path to ModelService handling sso user model>
}

USER_SERVICE should inherit from SsoUserMixin and AbstractModelService
and model should inherit from AbstractSsoUser
"""
from django.conf import settings
from rest_framework.authentication import TokenAuthentication

from w import utils
from w.services.technical.sso_service import SsoService
from w.tests.helpers import sso_test_helper


class SsoAuthBackend(TokenAuthentication):
    keyword = "Bearer"

    def _is_token_valid(self, token):
        return SsoService.is_token_valid(token)

    def _create_or_update_user(self, valid_token):
        user_service = utils.import_path(settings.SSO["USER_SERVICE"])
        data = valid_token.to_dict()
        uuid = data.pop("sso_uuid")
        user, created = user_service.create_or_update_by_uuid(uuid, data)
        return user.user

    def authenticate_credentials(self, token):
        # Checks token is active
        valid_token = self._is_token_valid(token)

        if "USER_SERVICE" not in settings.SSO:
            raise RuntimeError("'USER_SERVICE' is missing from SSO settings")

        user = self._create_or_update_user(valid_token)
        return user, None


class SsoAuthTestBackend(SsoAuthBackend):  # pragma: no cover
    _user = None

    @classmethod
    def set_user(cls, user):
        cls._user = user

    def _is_token_valid(self, token):
        if token != "invalid-token":
            return token
        with sso_test_helper.valid_token_failure():
            SsoService.is_token_valid(token)

    def _create_or_update_user(self, valid_token):
        return self._user



from dataclasses import dataclass, field
from typing import List

from w.mixins.dataclasses_mixin import DataclassMixin


@dataclass
class ImportRowError(DataclassMixin):
    num_row: int
    errors: List[str] = field(default_factory=list)


@dataclass
class ImportReport(DataclassMixin):
    nb_total: int = 0
    nb_imported: int = 0
    nb_ignored: int = 0
    nb_errors: int = 0
    errors: List[ImportRowError] = field(default_factory=list)



from django.contrib.auth.models import User
from django.db import models
from w.services.abstract_model_service import AbstractModelService
from w import exceptions


class UserService(AbstractModelService):
    _model = User

    @classmethod
    def get_or_create(cls, **attrs):
        # noinspection PyUnresolvedReferences
        try:
            return cls._model.objects.get(username=attrs.get("username"))
        except cls._model.DoesNotExist:
            return cls._model.objects.create(**attrs)

    @classmethod
    def create(cls, **attrs) -> models.Model:
        return User.objects.create_user(**attrs)

    @classmethod
    def get_by_email(cls, email):
        """
        Retrieve user by its email

        Returns:
            User
        """
        return cls._model.objects.get(email=email)

    @classmethod
    def check_by_email(cls, email):
        """
        Check user exists by its email

        if found return user else raise NotFoundError

        Raises
            NotFoundError
        """
        # noinspection PyUnresolvedReferences
        try:
            return cls.get_by_email(email)
        except cls._model.DoesNotExist:
            raise exceptions.NotFoundError(f"user not found (email={email})")



from django.core import mail

from w.services.technical.mail_service import EmailService
from w.tests.mixins.testcase_mixin import TestCaseMixin
from w.tests.serializers.serpy_serializers import MailOutboxSerializer


class TestEmailService(TestCaseMixin):

    # noinspection PyMethodMayBeStatic
    def setup_method(self, method):
        """clear mail outbox"""
        mail.outbox = []

    """
    send
    """

    def test_send_with_simple_email_return_int(self):
        """
        Ensure txt email sent succeed
        """
        params = {
            "recipients": "to@user.mail",
            "subject": "a mail subject",
            "message": "a mail content",
            "from_email": "from@user.com",
        }
        assert 1 == EmailService.send(**params)
        self.assert_equals_resultset(MailOutboxSerializer(mail.outbox[0]).data)

    def test_send_with_complex_email_return_int(self):
        """
        Ensure complex email (rendered via template) sent succeed
        """
        params = {
            "recipients": {
                "to": ["user1@to.com", "user2@to.com"],
                "bcc": "user1@bcc.com",
                "cc": ["user21@cc.com", "user22@cc.com"],
            },
            "subject": {
                "template_name": "email/email_subject.txt",
                "context": {"params1": "valueOne", "params2": "valueTwo"},
            },
            "message": {
                "template_name": "email/email_message.txt",
                "context": {"params1": "value11", "params2": "value22"},
            },
            "reply_to": "noreply@from.com",
        }
        assert 1 == EmailService.send(**params)
        self.assert_equals_resultset(MailOutboxSerializer(mail.outbox[0]).data)

    def test_send_with_html_email_return_int(self):
        """
        Ensure html email sent succeed
        """
        params = {
            "recipients": "to@user.mail",
            "subject": "a html email",
            "message": {
                "template_name": "email/email_message.html",
                "context": {"params1": "value11", "params2": "value22"},
            },
            "from_email": "from@user.com",
        }
        assert 1 == EmailService.send(**params)
        self.assert_equals_resultset(MailOutboxSerializer(mail.outbox[0]).data)



from contextlib import contextmanager
from unittest.mock import patch

from w.services.technical.yousign_service import YouSignService


def get_target(method_name):
    return YouSignService.get_patch_target(method_name)


@contextmanager
def create_success(return_value):
    with patch(get_target("create"), return_value=return_value) as m:
        yield m


@contextmanager
def create_yousign_failure():
    side_effect = RuntimeError("yousign api failed")
    with patch(get_target("create"), side_effect=side_effect) as m:
        yield m



class DrfValidatorSerializerMixin:
    def update(self, instance, validated_data):  # pragma: no cover
        raise RuntimeError("not intended to be used")

    def create(self, validated_data):  # pragma: no cover
        raise RuntimeError("not intended to be used")



import copy
from uuid import uuid4

from django.contrib.auth.models import User
from django.db import transaction

from w.services.technical.dict_service import DictService
from w.services.technical.list_service import ListService
from w.services.technical.sso_service import SsoService


class SsoUserMixin:
    _sso2django_remapping = {
        "email": "email",
        "username": "username",
        "firstName": "first_name",
        "lastName": "last_name",
    }

    @classmethod
    def remap_sso2django_user(cls, sso_data):
        return DictService.remap_keys(sso_data, cls._sso2django_remapping)

    @classmethod
    def _update_auth_user(cls, user, data):
        return cls.get_user_service().update(user, **data)

    @classmethod
    def _create_auth_user(cls, user_data):
        data = DictService.keep_keys(
            user_data, ["username", "email", "first_name", "last_name", "password"]
        )
        return cls.get_user_service().create(**data)

    @classmethod
    def create_from_sso(cls, uuid, user_data) -> User:
        """Create user (Django) and sso user from sso"""
        data = copy.deepcopy(user_data)
        if "password" not in user_data:
            data["password"] = str(uuid4())
        user = cls._create_auth_user(data)
        return cls.create(sso_uuid=uuid, user=user, list_apps=user_data["list_apps"])

    @classmethod
    def update_from_sso(cls, user, user_data):
        with transaction.atomic():
            list_apps = user_data.pop("list_apps")
            if ListService.are_different(user.list_apps, list_apps):
                cls.update(user, list_apps=list_apps)
            cls._update_auth_user(user.user, user_data)
        return user

    @classmethod
    def create_or_update_by_uuid(cls, uuid, user_data):
        """
        Create or update Django user from SSO server
        Args:
            uuid: sso uid
            user_data: {
                "username": "example",
                "first_name": "Example",
                "last_name": "Example",
                "email": "email@fake.com"
            }

        Returns:
            user(User), created(Bool)
        """
        qs = cls._model.objects.filter(sso_uuid=uuid)
        if qs.exists():
            return cls.update_from_sso(qs.first(), user_data), False
        return cls.create_from_sso(uuid, user_data), True

    @classmethod
    def get_by_uuid(cls, uuid):
        return cls._model.objects.get(sso_uuid=uuid)

    @classmethod
    def get_by_uuid_if_exist(cls, uuid):
        qs = cls._model.objects.filter(sso_uuid=uuid)
        if qs.exists():
            return qs.first()
        return None

    @classmethod
    def get_or_create_sso_user(cls, user: User, password: str, list_apps: list):
        attrs = DictService.keep_keys(
            cls.get_user_service().to_dict(user),
            ["email", "username", "first_name", "last_name"],
        )
        client_roles = {app: [f"{app}"] for app in list_apps}
        return SsoService.get_or_create_user(
            {**attrs, "password": password, "client_roles": client_roles}
        )

    @classmethod
    def get_user_service(cls):  # pragma: no cover
        """ get UserService on Django User model"""
        raise RuntimeError("need to be implemented")



from contextlib import contextmanager
from unittest.mock import patch

from w.services.technical.date_service import DateService


@contextmanager
def today_is(d):
    with patch("django.utils.timezone.now", return_value=DateService.to_datetime(d)):
        with patch("arrow.utcnow", return_value=DateService.get(d)) as m:
            yield m



from django.core.signing import BadSignature

from w.services.abstract_service import AbstractService


class AuthService(AbstractService):
    @classmethod
    def generate_token(cls):  # pragma: no cover
        # generate unique token
        import uuid

        return uuid.uuid4().hex

    @classmethod
    def generate_random_pwd(cls):
        return cls.generate_token()

    @classmethod
    def get_crypto_hash(cls, text: str, salt: str) -> str:
        """
        Returns:
            str: signature hash
        """
        from django.core.signing import Signer

        signer = Signer(salt)
        signed_text = signer.sign(text)
        return signed_text.replace(f"{text}:", "")

    @classmethod
    def is_crypto_hash_valid(cls, text: str, hash: str, salt: str) -> bool:
        """
        check signature (<text>:<crypto hash>) and return text on success
        else raise RuntimeError
        """
        from django.core.signing import Signer

        signer = Signer(salt)
        try:
            signer.unsign(f"{text}:{hash}")
            return True
        except BadSignature:
            return False



from w.services.abstract_service import AbstractService
from django.db import connection


class DbService(AbstractService):
    @staticmethod
    def exec_sql(sql, params=None):
        if params is None:
            params = []
        with connection.cursor() as cursor:
            cursor.execute(sql, params)
            columns = [col[0] for col in cursor.description]
            result = [dict(zip(columns, row)) for row in cursor.fetchall()]
        return result



from django.template.loader import render_to_string
from weasyprint import HTML
from weasyprint.fonts import FontConfiguration

from w.services.technical.filesystem_service import FilesystemService


class PdfService:
    @classmethod
    def write_file(cls, filename, content) -> bytes:
        """
        Create a PDF file

        Args:
            filename: string output file path
            content (str|dict): message or template (as dict) :
                {"template_name": <str>, "context": <dict> }

        Returns:
            bytes: pdf binary
        """

        if not content:
            return None

        font_config = FontConfiguration()

        # html content
        if isinstance(content, dict):
            content = render_to_string(**content)
        html = HTML(string=content)

        # create pdf content
        pdf = html.write_pdf(None, font_config=font_config)

        # create file
        FilesystemService.write_binary_file(filename, pdf)

        return pdf



from dataclasses import dataclass, field
from w.mixins.dataclasses_mixin import DataclassMixin

default_subject = "Hey! Vous êtes invité à signer un document"
default_message = (
    'Bonjour <tag data-tag-type="string" '
    'data-tag-name="recipient.firstname"></tag> <tag '
    'data-tag-type="string" data-tag-name="recipient.lastname"></tag>,'
    "<br><br> Vous êtes invité à signer un document, veuillez cliquer "
    'sur le bouton suivant pour le lire :<br><br><tag data-tag-type="button" '
    'data-tag-name="url" data-tag-title="Accéder au contrat">Accéder '
    "au document</tag>"
)


@dataclass
class YouSignMail(DataclassMixin):
    subject: str = field(default_factory=lambda: default_subject)
    message: str = field(default_factory=lambda: default_message)
    to: list = field(default_factory=lambda: ["@member"])


@dataclass
class YouSignRequest(DataclassMixin):
    procedure_name: str
    procedure_desc: str
    filename: str
    member_firstname: str
    member_lastname: str
    member_email: str
    member_phone: str
    file_signature_position: str
    member_started_mail: YouSignMail = field(default_factory=YouSignMail)

    def get_procedure_payload(self, file_id):
        return {
            "name": self.procedure_name,
            "description": self.procedure_desc,
            "members": [
                {
                    "firstname": self.member_firstname,
                    "lastname": self.member_lastname,
                    "email": self.member_email,
                    "phone": self.member_phone,
                    "fileObjects": [
                        {
                            "file": file_id,
                            "page": 1,
                            "position": self.file_signature_position,
                        }
                    ],
                }
            ],
            "config": {
                "email": {"member.started": [self.member_started_mail.to_dict()]}
            },
        }



import os

from w.services.technical.models.yousign import YouSignRequest
from w.services.technical.yousign_service import YouSignService

YOUSIGN_API_URL = "https://staging-api.yousign.com"
YOUSIGN_API_KEY = "330fdd14a571dd22241ad8002f2ab545"

pdf = f"{os.path.dirname(os.path.abspath(__file__))}/../pdf/exemple_pdf.pdf"
data = {
    "procedure_name": "Test de création",
    "procedure_desc": "Un test pour valider le fonctionnement",
    "filename": pdf,
    "member_firstname": "François",
    "member_lastname": "Schneider",
    "member_email": "fr-schneider@orange.fr",
    "member_phone": "+33662344601",
}

yousign_request = YouSignRequest(**data)

if __name__ == "__main__":
    YouSignService.init(YOUSIGN_API_URL, YOUSIGN_API_KEY)
    member_id = YouSignService.create(yousign_request)
    print(f"member_id = {member_id}")

# /members/653285fa-0aee-44c2-a065-ee105b2090de



from dataclasses import dataclass
from datetime import datetime

from w.drf.mixins import (
    ListModelServiceMixin,
    RetrieveModelServiceMixin,
    CreateModelServiceMixin,
    UpdateModelServiceMixin,
    DeleteModelServiceMixin,
)
from w.drf.viewsets import ViewSet, ModelServiceViewSet
from w.mixins.dataclasses_mixin import DataclassMixin
from w.tests.fixtures.datasets.django_app import serpy_serializer
from w.tests.fixtures.datasets.django_app import drf_serializer
from w.tests.fixtures.datasets.django_app.example_service import (
    ExampleService,
)


@dataclass
class Simple(DataclassMixin):
    integer: int
    string: str
    date: datetime.date


class SimpleViewset(ViewSet):
    serializers = {
        "default": serpy_serializer.SimpleSerializer,
        "create_validation": drf_serializer.SimpleCreateValidation,
    }

    def create(self, request):
        validated_data = self.check_is_valid()
        return self.get_post_response(Simple(**validated_data))


class ModelViewset(
    DeleteModelServiceMixin,
    UpdateModelServiceMixin,
    CreateModelServiceMixin,
    RetrieveModelServiceMixin,
    ListModelServiceMixin,
    ModelServiceViewSet,
):
    serializers = {
        "default": serpy_serializer.ExampleSerializer,
        "create_validation": drf_serializer.ExampleCreateValidation,
        "update_validation": drf_serializer.ExampleUpdateValidation,
    }
    service = ExampleService



from unittest.mock import patch, Mock

from w.services.technical.auth_service import AuthService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestAuthService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.text = "mon text à signer"
        cls.signed_text = (
            "mon text à signer:HkmUAVUQI6KNcQjbQX1ZZGfJWMlp5NCRNqOXwwQTaUs"
        )
        cls.salt = "mon sel"
        cls.hash = "HkmUAVUQI6KNcQjbQX1ZZGfJWMlp5NCRNqOXwwQTaUs"

    """
    generate_random_pwd
    """

    def test_generate_random_pwd_with_success_return_str(self):
        """Ensure method return random password"""
        with patch("uuid.uuid4") as uuid4:
            uuid4.return_value = Mock(hex="uuid-token")
            assert AuthService.generate_random_pwd() == "uuid-token"

    """
    get_crypto_hash
    """

    def test_get_crypto_hash_with_success_return_str(self):
        """Ensure method succeeds"""
        actual = AuthService.get_crypto_hash(self.text, salt=self.salt)
        assert actual == self.hash

    """
    is_crypto_hash_valid
    """

    def test_is_crypto_hash_valid_with_wrong_hash_return_false(self):
        """Ensure method raise return false with wrong hash"""
        actual = AuthService.is_crypto_hash_valid(
            self.text, "mauvaishash", salt=self.salt
        )
        assert actual is False

    def test_is_crypto_hash_valid_with_wrong_salt_return_false(self):
        """Ensure method raise return false with wrong salt"""
        actual = AuthService.is_crypto_hash_valid(
            self.text, self.hash, salt="mauvais sel"
        )
        assert actual is False

    def test_is_crypto_hash_valid_with_success_return_true(self):
        """Ensure method succeed"""
        actual = AuthService.is_crypto_hash_valid(self.text, self.hash, self.salt)
        assert actual is True



from contextlib import contextmanager

from w.services.technical.auth_service import AuthService
from w.tests.helpers import service_test_helper


def get_generate_token_mock_config(token):
    return {
        "service": AuthService,
        "method_name": "generate_token",
        "return_value": token,
    }


@contextmanager
def mock_generate_token(token):
    mock = get_generate_token_mock_config(token)
    with service_test_helper.mock_service(**mock) as m:
        yield m



from django.conf import settings
from w.services.technical.google_map_service import GoogleMapService
from w.tests.mixins.testcase_mixin import TestCaseMixin

TEST_URL = "ma grosse adresses avec # des caratères à la con !$"
TEST_PLACE_ID = "ChIJN1t_tDeuEmsRUsoyG83frY4"


class TestGMapService(TestCaseMixin):
    def test_build_geocode_url(self):
        url = GoogleMapService.build_geocode_url(TEST_URL)
        assert (
            url == "https://maps.googleapis.com/maps/api/geocode/json?address="
            "ma+grosse+adresses+avec+%23+des+carat%C3%A8res+%C3%A0+la+con+"
            "%21%24&key=AIzaSyAYU9lXqmmetI3s7feQi_Yc6_f7KGvVPz4"
        )

    def test_geocode_service(self):
        resp = GoogleMapService.geocode_address(
            "Royal Albert Hall, Kensington Gore, Londres, Royaume-Uni"
        )
        if settings.MOCK_GMAP_CALLS:
            assert resp["lat"] == 10
            assert resp["lng"] == 10
        else:
            assert resp["lat"] == 51.5009088
            assert resp["lng"] == -0.177366

    def test_place_autocomplete(self):
        url = GoogleMapService.build_autocomplete_place_url(TEST_URL, "mysession_token")
        assert (
            url == "https://maps.googleapis.com/maps/api/place/queryautocomplete/json?"
            "input=ma+grosse+adresses+avec+%23+des+carat%C3%A8res+%C3%A0+la+con+%21%24"
            "&key=AIzaSyAYU9lXqmmetI3s7feQi_Yc6_f7KGvVPz4&langage=fr&components=country"
            "%3Afr%7Ccountry%3Ach%7Ccountry%3Abe%7Ccountry%3Ade%7Ccountry%3Aes"
            "&sessiontoken=mysession_token"
        )

    def test_autocomplete_address(self):
        resp = GoogleMapService.autocomplete_address("rue du mascaret, Bordeaux")
        if settings.MOCK_GMAP_CALLS:
            assert resp["location"]["lat"] == 10
            assert resp["location"]["lng"] == 10
        else:
            assert (
                resp["formatted_address"] == "Rue du Mascaret, 33800 Bordeaux, France"
            )
            assert resp["location"]["lat"] == 44.8262542
            assert resp["location"]["lng"] == -0.5518803

    def test_build_place_detail_address(self):
        url = GoogleMapService.build_place_detail_url(TEST_PLACE_ID)
        assert (
            url == "https://maps.googleapis.com/maps/api/place/details/json?"
            "place_id=ChIJN1t_tDeuEmsRUsoyG83frY4&key=AIzaSyAYU9lXqmmetI3s7feQi_Yc6_"
            "f7KGvVPz4&fields=geometry,formatted_address,place_id"
        )

    def test_get_place_detail(self):
        resp = GoogleMapService.get_place_detail(TEST_PLACE_ID)
        self.assert_equals_resultset(resp)



import errno
import os
import shutil
from pathlib import Path
from typing import Union

from w.services.abstract_service import AbstractService
from w.services.technical.date_service import DateService


class FilesystemService(AbstractService):
    """Service Filesystem"""

    @classmethod
    def is_dir_exists(cls, directory) -> bool:
        """
        Is it refers to an existing directory

        Args:
            directory(str):

        Returns:
            bool
        """
        return os.path.isdir(directory)

    @classmethod
    def is_file_exists(cls, filename) -> bool:
        """
        Is it refers to an existing file

        Args:
            filename(str):

        Returns:
            bool
        """
        return os.path.isfile(filename)

    @classmethod
    def check_dir_exists(cls, directory) -> None:
        """
        Check if a directory exists

        Args:
            directory(str):

        Returns:
            None if directory exists

        Raises:
            RuntimeError if directory does not exists
        """
        if cls.is_dir_exists(directory):
            return None
        raise RuntimeError(f"{directory} does not exists")

    @classmethod
    def check_file_exists(cls, filename) -> None:
        """
        Check if a file exists

        Args:
            filename(str):

        Returns:
            None if file exists

        Raises:
            RuntimeError if file does not exists
        """
        if cls.is_file_exists(filename):
            return None
        raise RuntimeError(f"{filename} does not exists")

    @classmethod
    def copy_dir(cls, src, dst) -> None:
        """
        Copy directory tree and overwrite existing files

        Args:
            src:
            dst:
        """
        cls.check_dir_exists(src)
        for root, dirs, files in os.walk(src):
            for file in files:
                rel_path = root.replace(src, "").lstrip(os.sep)
                dest_path = os.path.join(dst, rel_path)
                if not os.path.exists(dest_path):
                    os.makedirs(dest_path)
                shutil.copy(os.path.join(root, file), os.path.join(dest_path, file))

    @classmethod
    def copy_file(cls, src, dst, overwrite=False) -> None:
        """
        Copy file in a new location

        Args:
            src(str): file path ( ex : "/foo/bar/file.txt" )
            dst(str): new file path ( ex : "/foo/bar/file.txt" )
            overwrite(bool): force overwriting existing file
        """
        cls.check_file_exists(src)
        dest_path = os.path.dirname(dst)
        if not os.path.exists(dest_path):
            os.makedirs(dest_path)
        if not overwrite:
            if not cls.is_file_exists(dst):
                shutil.copy(src, dst)
        else:
            shutil.copy(src, dst)

    @classmethod
    def empty_dir(cls, directory) -> None:
        """
        Empty a directory

        Args:
            directory(str):

        Returns:
            None

        Raises:
            RuntimeError if directory does not exists
        """
        cls.check_dir_exists(directory)
        for root, dirs, files in os.walk(directory, topdown=False):
            for file in files:
                os.remove(os.path.join(root, file))
            if root != directory:
                os.rmdir(root)

    @classmethod
    def write_file(cls, filename, content):
        """
        Write content into file, erase it if already exists
        Args:
            filename(str): full path
            content(str): text content to write

        Returns:

        """
        cls.check_dir_exists(os.path.dirname(filename))
        with open(filename, "w") as f:
            f.write(content)

    @classmethod
    def write_binary_file(cls, filename, content):
        """
        Write content into file, erase it if already exists
        Args:
            filename(str): full path
            content(bites): text content to write

        Returns:

        """
        cls.check_dir_exists(os.path.dirname(filename))
        with open(filename, "wb") as f:
            f.write(content)

    @classmethod
    def read_file(cls, filename):
        """
        Read file
        Args:
            filename(str): full path

        Returns:
            str
        """
        cls.check_file_exists(filename)
        with open(filename) as f:
            return f.read()

    @staticmethod
    def check_file_can_be_created(filename) -> None:
        """
        create missing directories for filename

        Args:
            filename (str) : fullpath filename

        Returns:
            None
        """
        if not os.path.exists(filename):
            try:
                return os.makedirs(os.path.dirname(filename))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

    @staticmethod
    def get_path(path) -> Path:
        """ Get Path instance if path is not """
        if isinstance(path, Path):
            return path
        return Path(path)

    @classmethod
    def get_basename(cls, filename):
        """
        Get string representing the final path component, excluding the drive and root

        Args:
            filename(str|Path): filename

        Returns:
            str: string representing the final path component
        """
        return cls.get_path(filename).name

    @classmethod
    def has_suffix(cls, filename, suffix: Union[str, list]) -> bool:
        """
        Check if filename has suffix(es)

        Args:
            filename(str):
            suffix(str|list): suffix or list of suffixes

        Returns:
            bool
        """
        filename = cls.get_path(filename)
        if not isinstance(suffix, list):
            suffix = [suffix]
        return filename.suffix in suffix

    @staticmethod
    def prefix_filename_to_now(filename):
        """
        Add timestamp as prefix to the provided filename.
        Timestamp is set to current date and time.

        Args:
            filename(str):

        Returns:
            filename with prefix
        """
        return f"{DateService.to_str('YYYYMMDD_HHmmss')}-{filename}"



import factory
from faker.generator import random

from w.django.tests.factories.auth_factories import UserFactory
from w.tests.fixtures.datasets.django_app import models


class SsoUserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.SsoUser

    sso_uuid = factory.Sequence(lambda n: f"fake-sso-uuid-{n}")
    list_apps = factory.LazyFunction(lambda: random.choices(["app1", "app2", "app3"]))
    user = factory.SubFactory(UserFactory)



import datetime
import json
from decimal import Decimal

from w.services.abstract_service import AbstractService
from w.services.technical.filesystem_service import FilesystemService


class JsonService(AbstractService):
    @classmethod
    def dump_to_file(cls, data, filename, human_readable=False):
        """
        Dump json into file
        """
        default_args = {"obj": data}
        human_readable_args = {
            "indent": 4,
            "separators": (",", ": "),
            "default": str,
            # disable non-ASCII characters escape with \uXXXX sequences
            "ensure_ascii": False,
        }
        with open(filename, "w") as f:
            args = {**default_args, **{"fp": f}}
            if human_readable:
                args = {**args, **human_readable_args}
            json.dump(**args)

    @classmethod
    def to_json_dumpable(cls, data):
        """
        Convert data to be json dumpable
        """
        from w.services.technical.date_service import DateService

        if isinstance(data, Decimal):
            return float(data)

        if isinstance(data, datetime.datetime):
            return DateService.to_mysql_datetime(data)

        if isinstance(data, datetime.date):
            return DateService.to_mysql_date(data)

        if isinstance(data, bytes):
            return data.decode("utf-8")

        if isinstance(data, dict):
            for k, v in data.items():
                data[k] = cls.to_json_dumpable(v)
            return data

        if isinstance(data, list):
            for idx, i in enumerate(data):
                data[idx] = cls.to_json_dumpable(i)
            return data

        if isinstance(data, tuple):
            data = cls.to_json_dumpable(list(data))

        if isinstance(data, set):
            data = cls.to_json_dumpable(list(data))

        return data

    @staticmethod
    def load_from_file(filename):
        """
        Load json from file

        Args:
            filename(str):

        Returns:
            Any
        """
        FilesystemService.check_file_exists(filename)
        with open(filename) as f:
            return json.load(f)

    @staticmethod
    def load_from_str(str_json: str):
        """
        Load json from str

        Args:
            str_json(str):

        Returns:
            Any
        """
        return json.loads(str_json)

    @staticmethod
    def is_valid(json_data) -> bool:
        """
        Check json is valid

        Args:
            json_data(str): json to check

        Returns:
            bool
        """
        try:
            json.loads(json_data)
            return True
        except ValueError:
            return False

    @classmethod
    def dump(cls, data) -> str:
        """
        Dump data into json string, convert data to be dumpable

        Args:
            data(mixed): python data

        Returns:
            str: json
        """
        return json.dumps(JsonService.to_json_dumpable(data))



import os
from pathlib import Path
import pytest

from w.services.technical.filesystem_service import FilesystemService
from w.tests.helpers import date_test_helper
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestFilesystemService(TestCaseMixin):
    """Test unit suite for FilesystemService"""

    def setup_method(self):
        FilesystemService.clear()
        self.clean_sandbox()

    """
    check_dir_exits
    """

    def test_check_dir_exists_with_unknown_dir_raise_runtime_error(self):
        """ Ensure check dir raise RuntimeError if dir does not exists """
        with pytest.raises(RuntimeError, match="unknown-dir does not exists"):
            FilesystemService.check_dir_exists("unknown-dir")

    def test_check_dir_exists_with_success_return_none(self):
        """ Ensure check dir succeed """
        assert not FilesystemService.check_dir_exists(
            self.get_datasets_dir("filesystem")
        )

    """
    check_file_exits
    """

    def test_check_file_exists_with_unknown_file_raise_runtime_error(self):
        """ Ensure check file raise RuntimeError if file does not exists """
        with pytest.raises(RuntimeError, match="unknown-file does not exists"):
            FilesystemService.check_file_exists("unknown-file")

    def test_check_file_exists_with_success_return_none(self):
        """ Ensure check file succeed """
        assert not FilesystemService.check_file_exists(
            self.get_datasets_dir("filesystem/dir1/file1.txt")
        )

    """
    copy_dir
    """

    def test_copy_dir_with_unknown_source_raise_runtime_error(self):
        """ Ensure copy dir raise RuntimeError if dir does not exists """
        with pytest.raises(RuntimeError, match="unknown-dir does not exists"):
            FilesystemService.copy_dir("unknown-dir", "dest")

    def test_copy_dir_with_success_return_none(self):
        """ Ensure copy dir succeed """
        assert not FilesystemService.copy_dir(
            self.get_datasets_dir("filesystem"), self.get_sandbox_dir()
        )
        self.assert_equals_resultset(
            sorted(
                [
                    str(i).replace(self.get_sandbox_dir(), ".")
                    for i in Path(self.get_sandbox_dir()).glob("**/*")
                ]
            )
        )

    """
    empty_dir
    """

    def test_empty_dir_with_unknown_dir_raise_runtime_error(self):
        """ Ensure empty dir raise RuntimeError if dir does not exists """
        with pytest.raises(RuntimeError, match="unknown-dir does not exists"):
            FilesystemService.empty_dir("unknown-dir")

    def test_empty_dir_with_success_return_none(self):
        """ Ensure empty dir succeed """
        FilesystemService.copy_dir(
            self.get_datasets_dir("filesystem"), self.get_sandbox_dir()
        )
        assert FilesystemService.empty_dir(self.get_sandbox_dir()) is None
        assert FilesystemService.is_dir_exists(self.get_sandbox_dir())
        assert not FilesystemService.is_dir_exists(self.get_sandbox_dir("dir1"))

    """
    write_file
    """

    def test_write_file_with_unknown_dir_raise_runtime_error(self):
        """ Ensure writing file raise RuntimeError if dest dir does not exists """
        with pytest.raises(RuntimeError, match="unknown-dir does not exists"):
            FilesystemService.write_file("unknown-dir/file.txt", "raise error")

    def test_write_file_with_success_return_none(self):
        """ Ensure writing file succeed """
        self.clean_sandbox()
        filename = self.get_sandbox_dir("writing_success.txt")
        FilesystemService.write_file(filename, "so successfull !!")
        assert os.path.exists(filename)
        with open(filename) as f:
            assert "so successfull !!" == f.read()

    """
    read_file
    """

    def test_read_file_with_unknown_dir_raise_runtime_error(self):
        """ Ensure reading file raise RuntimeError if file does not exists """
        filename = self.get_datasets_dir("filesystem/dir1/fileUnknown.txt")
        with pytest.raises(RuntimeError, match=f"{filename} does not exists"):
            FilesystemService.read_file(filename)

    def test_read_file_with_success_return_none(self):
        """ Ensure reading file succeed """
        self.clean_sandbox()
        filename = self.get_datasets_dir("filesystem/dir1/file1.txt")
        actual = FilesystemService.read_file(filename)
        assert actual == "file1"

    """
    check_file_can_be_created
    """

    def test_check_file_can_be_created_with_success_return_none(self):
        """ Ensure missing directories will be created """
        self.clean_sandbox()
        filename = self.get_sandbox_dir("filesystem/dir1/file1.txt")
        assert FilesystemService.check_file_can_be_created(filename) is None
        assert FilesystemService.is_dir_exists(self.get_sandbox_dir("filesystem/dir1"))

    """
    prefix_filename_to_now
    """

    def test_prefix_filename_to_now_with_success_return_dated_filename(self):
        """ Ensure filename is returned with a prefix set to today's date """
        with date_test_helper.today_is("2021-05-06 12:34:56"):
            assert (
                FilesystemService.prefix_filename_to_now("my_file.txt")
                == "20210506_123456-my_file.txt"
            )



from w.services.abstract_model_service import AbstractModelService
from w.tests.fixtures.datasets.django_app import models


class ExampleService(AbstractModelService):
    _model = models.ExampleWithInternalDependencies



from rest_framework import serializers


class StrictBooleanField(serializers.BooleanField):
    def to_internal_value(self, data):
        if data == "":
            data = None
        if data in ("true", "True", "1"):
            data = True
        if data in ("false", "False", "0"):
            data = False

        if self.required and data is None:
            raise serializers.ValidationError("This field may not be blank.")

        if data not in (False, True):
            raise serializers.ValidationError("This field must be a boolean.")

        return data



import os
from django.core.mail import EmailMessage
from django.template.loader import render_to_string
from django.conf import settings

from w.services.abstract_service import AbstractService


class EmailService(AbstractService):
    @staticmethod
    def render_template(template_name, context=None):
        """
        Render mail template
        """
        return render_to_string(template_name, context)

    @classmethod
    def send(cls, recipients, subject, message, attach_files=None, **options):
        """
        Send an email

        Args:
            recipients(str|list|dict): email or list of emails or dict
                {
                  "to": <email or list of emails>
                  "bcc": <email or list of emails>
                  "cc": <email or list of emails>
                }
            subject (str|dict): subject or template (as dict) :
                {"template_name": <str>, "context": <dict> }
            message (str|dict): message or template (as dict) :
                {"template_name": <str>, "context": <dict> }
            attach_files:  (dict) list of file to attach
                {"attach_files": [<str>], "context": <dict> }
            **options (dict):
                    'from_email': <from_email, default is settings.DEFAULT_FROM_EMAIL>,
                    'reply_to' : <reply to email, None by default>

        Returns:
            int: number of email sent
        """
        default_params = {
            "from_email": settings.DEFAULT_FROM_EMAIL,
            "to": None,
            "bcc": None,
            "cc": None,
            "connection": None,
            "attachments": None,
            "headers": None,
            "reply_to": settings.DEFAULT_REPLY_TO,
        }

        if not isinstance(recipients, dict):
            recipients = {"to": recipients}

        # noinspection PyDictCreation
        params = {**default_params, **recipients, **options}

        # convert str recipient to list
        for key in ["to", "bcc", "cc", "reply_to"]:
            if params[key] and not isinstance(params[key], list):
                params[key] = [params[key]]

        if isinstance(subject, dict):
            subject = render_to_string(**subject)

        is_html = False
        if isinstance(message, dict):
            _, ext = os.path.splitext(message["template_name"])
            is_html = ext == ".html"
            message = render_to_string(**message)

        params["subject"] = subject
        params["body"] = message
        email = EmailMessage(**params)

        if is_html:
            email.content_subtype = "html"

        if attach_files:
            for file in attach_files:
                email.attach_file(file)

        return email.send()



from rest_framework_simplejwt.authentication import (
    JWTAuthentication as SimpleJWTAuthentication,
)


class JWTAuthentication(SimpleJWTAuthentication):
    def authenticate(self, request):
        header = self.get_header(request)
        if header is None:
            return None

        raw_token = self.get_raw_token(header)
        if raw_token is None:
            return None

        validated_token = self.get_validated_token(raw_token)
        return self.get_user(validated_token), validated_token



from django.db import connection
from factory import SubFactory


class FactoryBoyMixin:
    """ Mixin for FactoryBoy """

    # factories list
    _factories = {}

    @classmethod
    def factory_reset_sequence(cls, facto, value=None):
        if value is None:
            value = 1
        # as we cannot reset descendant factories, reset parent
        # noinspection PyProtectedMember
        facto._meta.counter_reference.factory.reset_sequence(value)
        for attr, dec in facto._meta.declarations.items():
            if isinstance(dec, SubFactory):
                cls.factory_reset_sequence(dec.factory_wrapper.factory)

    @classmethod
    def reset_sequence(cls, value=1, reset_pk_model=False):
        value_not_model = value if value else 1
        if reset_pk_model:
            cls.reset_all_db_sequences()
        for name, factory in cls._factories.items():
            cls.factory_reset_sequence(factory, value_not_model)

    @classmethod
    def reset_all_db_sequences(cls):
        if connection.vendor != "postgresql":
            raise RuntimeError("connection not postgresql. reset_all canceled")

        sql = "SELECT c.relname FROM pg_class c WHERE c.relkind = 'S';"
        with connection.cursor() as c:
            c.execute(sql)
            results = c.fetchall()

            for sequence in results:
                sql = f"ALTER SEQUENCE {sequence[0]} RESTART 1;"
                c.execute(sql)

    @classmethod
    def _check_factory_exists(cls, factory_key):
        if factory_key not in cls._factories:
            raise RuntimeError(
                f"{factory_key} missing, add it to {__class__}._factories"
            )

    @classmethod
    def create(cls, factory_key, nb=1, **kwargs):
        cls._check_factory_exists(factory_key)
        return (
            cls._factories[factory_key](**kwargs)
            if nb == 1
            else cls._factories[factory_key].create_batch(nb, **kwargs)
        )

    @classmethod
    def build_as_dict(cls, factory_key, **kwargs):
        cls._check_factory_exists(factory_key)
        return cls._factories[factory_key].stub(**kwargs).__dict__



from w.services.technical.list_service import ListService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestListService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        cls.list1 = [10, 11, 12, 13, 14, 16, 15]
        cls.list2 = [10, 11, 12, 18, 19, 16]

    """
    list_differences
    """

    def test_list_differences_with_no_difference_return_list(self):
        """Ensure method succeed with no difference"""
        assert ListService.list_differences(["no", "diff"], ["no", "diff"]) == []

    def test_list_differences_with_success_return_list(self):
        """Ensure method succeed with no difference"""
        expected = [18, 19, 13, 14, 15]
        assert ListService.list_differences(self.list1, self.list2) == expected

    """
    are_same
    """

    def test_are_same_with_no_difference_return_true(self):
        """ Ensure method return True if lists are same """
        assert ListService.are_same(["no", "diff"], ["no", "diff"]) is True

    def test_are_same_with_success_return_false(self):
        """Ensure method return False if lists are different"""
        assert ListService.are_same(self.list1, self.list2) is False

    """
    are_different
    """

    def test_are_different_with_no_difference_return_false(self):
        """ Ensure method return False if lists are same """
        assert ListService.are_different(["no", "diff"], ["no", "diff"]) is False

    def test_are_different_with_success_return_true(self):
        """Ensure method return True if lists are different"""
        assert ListService.are_different(self.list1, self.list2) is True



#!/usr/bin/env python
import os
import setuptools

if __name__ == "__main__":
    version = os.environ.get("WEDGELIB_VERSION", "0.0.0")
    setuptools.setup(
        version=version,
    )
    if version == "0.0.0":
        print(
            "!!! Attention !!! \n"
            "La variable d'environnment WEDGELIB_VERSION n'est pas définie. "
            "Le build est configuré pour fonctionner automatiquement sur la création "
            "d'une release dans GitHub.\n"
            "NE PAS PUBLIER SUR PYPI CE BUILD"
        )



import copy
import uuid

from django.core.exceptions import ObjectDoesNotExist, FieldDoesNotExist
from django.db import models
from factory import SubFactory

from w.tests.helpers import date_test_helper


class DataTestFactory:
    def __init__(self):
        self.data_repository = dict()
        self.factory_repository = list()
        self._today_is = "2015-09-03 11:15:00"

    def _factory_reset_sequence(self, facto, value=None):
        if value is None:
            value = 1
        # as we cannot reset descendant factories, reset parent
        # noinspection PyProtectedMember
        facto._meta.counter_reference.factory.reset_sequence(value)
        for attr, dec in facto._meta.declarations.items():
            if isinstance(dec, SubFactory):
                self._check_and_reset_factory(dec.factory_wrapper.factory)

    def _check_and_reset_factory(self, factory_class):
        if factory_class not in self.factory_repository and factory_class is not None:
            self._factory_reset_sequence(factory_class)
            self.factory_repository.append(factory_class)

    @staticmethod
    def _find_direct_relation_name(fk_model, parent_model):
        return [
            f.name
            for f in parent_model._meta.fields
            if f.many_to_one is True and f.related_model == fk_model
        ][0]

    def _create_fk(self, data_def):
        # si on trouve une relation directe, on crée l'objet lié cash pistache,
        # sauf si c'est un objet existant un objet existant est référencé par le
        # mot clef _data_ref
        use_data_ref = data_def.get("data_ref", None)
        if use_data_ref in self.data_repository and len(data_def) >= 2:
            raise RuntimeError(
                "l'entrée " + use_data_ref + " est déjà définie par ailleurs."
            )

        if use_data_ref and len(data_def) == 1:
            # si l'entrée dataref est trouvée et qu'elle est seule, on va faire
            # une recherche dans le repo de data préalablement crée.
            # @TBD : si la relation n'est pas trouvé, on pourrait déférrer la
            # création à plus tard, quand l'objet serait trouvé
            return self.get_by_ref(use_data_ref)

        return self._load_data(data_def)
        # on pourra ensuite passer cet objet à l'objet en cours pour création

    @staticmethod
    def _is_many2many_field(parent_model: models.Model, fieldname):
        try:
            field_class = parent_model._meta.get_field(fieldname).__class__.__name__
            return field_class in ["ManyToManyField", "ManyToManyRel"]
        except FieldDoesNotExist:
            raise RuntimeError(
                f"{fieldname} not found in {parent_model.__class__.__name__}"
            )

    def _prepare_relations(self, data_def):
        fk_relations = {}
        many2many_relations = {}
        for key, value in data_def["values"].copy().items():
            # on va examiner les relations
            if isinstance(value, dict):
                # si on a un élément de type dict, il s'agit d'une relation directe
                fk_relations[key] = self._create_fk(value)
                continue

            if isinstance(value, list):
                # si on a un élément de type list, alors c'est une relation inverse ou
                # many2many. Il faut créer l'objet en cours avant de créer les objets
                # liés on pop donc la structure des enfants,
                # pour permettre la création de l'objet en cours, sans être pollué par
                # les enfants
                many2many_relations[key] = data_def["values"].pop(key)

        return fk_relations, many2many_relations

    def _create_by_factory(self, data_def, fieldname=None, parent_model=None):
        factory_class = data_def.get("factory")
        self._check_and_reset_factory(factory_class)

        # prepare relations => create needed fk or delay many2many relations creation
        fk_relations, many2many_relations = self._prepare_relations(data_def)

        create_datas = {**data_def["values"]}

        # si on a un objet en fk, il faut l'ajouter dans les paramètres de la factory,
        # via le nom de la relation, qu'il nous faut trouver
        if parent_model and not self._is_many2many_field(parent_model, fieldname):
            # la relation est en réalité une relation foreign key inverse
            rel_name = self._find_direct_relation_name(
                parent_model._meta.model, factory_class._meta.model
            )
            create_datas.update({rel_name: parent_model})

        if fk_relations:
            create_datas.update(**fk_relations)

        created = factory_class(**create_datas)

        # si on a poppé des relations inverses, il faut alors passer l'objet crée, à
        # l'étage du dessous pour création car certaines sont non nullables
        if many2many_relations:
            for fieldname, value in many2many_relations.items():
                self._load_data(value, fieldname=fieldname, related_object=created)

        return created

    def _parse_and_create_object(self, data_def, fieldname=None, parent_model=None):
        # si l'objet à créer est un callable, on se pose pas de question, on call
        obj_to_create = copy.deepcopy(data_def)
        use_data_ref = obj_to_create.pop("data_ref", None)

        if "factory" in obj_to_create:
            data = self._create_by_factory(obj_to_create, fieldname, parent_model)
        elif use_data_ref:
            data = self.get_by_ref(use_data_ref)
        else:
            raise RuntimeError(
                f"factory not found for {obj_to_create} or missing data_ref"
            )

        if parent_model and self._is_many2many_field(parent_model, fieldname):
            getattr(parent_model, fieldname).add(data)

        return data

    def _add_to_data_repository(self, data_ref=None, data_content=None):
        if data_ref is None:
            data_ref = uuid.uuid4()
        self.data_repository.update({data_ref: data_content})

    def get_by_ref(self, data_ref):
        found = self.data_repository.get(data_ref, None)
        if found is None:
            raise ObjectDoesNotExist(
                "dataref : '" + data_ref + "' Not previously created !"
            )
        return found

    @staticmethod
    def _get_unique_data_ref(data_ref, nb):
        # nous devons assurer l'unicité de la dataref dans le cas
        # d'instanciation multiple
        return f"{data_ref}_{nb}" if data_ref and nb > 1 else data_ref

    @staticmethod
    def _get_from_db(model, selector):
        return model.objects.get(**selector)

    def _load_data_def_context(self, data_def) -> None:
        """ Load context if necessary """
        for context_data_def in data_def.get("context", []):
            self._load_data(context_data_def)

    def _load_dict_data(self, data_def, many2many_key=None, fk_object=None):
        # load some context first if necessary
        self._load_data_def_context(data_def)

        # load from db
        db_data = data_def.pop("from_db", None)
        if db_data:
            return self._get_from_db(**db_data)

        data_occurrence = data_def.pop("nb", 1)
        data_ref = data_def.pop("ref", None)
        created_objects = []
        for cpt in range(0, data_occurrence):
            # on doit travailler sur une copie, pour pouvoir pop tranquillement les
            # valeurs spéciales, plusieurs fois.
            created_object = self._parse_and_create_object(
                data_def, many2many_key, fk_object
            )
            unique_data_ref = self._get_unique_data_ref(data_ref, cpt + 1)
            self._add_to_data_repository(unique_data_ref, created_object)
            created_objects.append(created_object)
        return created_objects if data_occurrence > 1 else created_objects[0]

    def _load_list_data(self, data_def, many2many_key, fk_object):
        created = []
        for data in data_def:
            created_object = self._load_data(data, many2many_key, fk_object)
            created.append(created_object)
        return created

    def _load_data(self, data_def, fieldname=None, related_object=None):
        # besoin de faire une copie dans le cas ou les définitions de datas
        # sont utilisées plusieurs fois
        data_def_copy = data_def.copy()
        if isinstance(data_def_copy, list):
            return self._load_list_data(data_def_copy, fieldname, related_object)

        if "values" not in data_def_copy:
            # add missing data key
            data_def_copy["values"] = {}

        return self._load_dict_data(data_def_copy, fieldname, related_object)

    def today_is(self, today_date: str) -> "DataTestFactory":
        """
        Force today's date

        Args:
            today_date(str): YYYY-MM-DD HH:MM:SS

        Returns:
            DataTestFactory
        """
        self._today_is = today_date
        return self

    def build(self, data_def, nb=None):
        with date_test_helper.today_is(self._today_is):
            data_def_copy = data_def.copy()
            if nb and nb > 0:
                data_def_copy["nb"] = nb
            return self._load_data(data_def_copy)



from django.contrib.auth.models import User
from django.db import models


class TextChoices(models.TextChoices):
    @classmethod
    def list_codes(cls):
        return [code for code, label in cls.choices]


class AbstractCreatedUpdatedModel(models.Model):
    created_at = models.DateTimeField("crée le", auto_now_add=True)
    updated_at = models.DateTimeField("maj le", auto_now=True)

    class Meta:
        abstract = True


class AbstractSsoUser(AbstractCreatedUpdatedModel, models.Model):
    sso_uuid = models.CharField(max_length=100, db_index=True)
    list_apps = models.JSONField()
    user = models.OneToOneField(User, related_name="sso_user", on_delete=models.CASCADE)

    class Meta:
        abstract = True


class Reference(AbstractCreatedUpdatedModel):
    code = models.CharField(max_length=20, primary_key=True)
    label = models.CharField(max_length=100)

    class Meta:
        abstract = True
        ordering = ["label"]

    def __str__(self):
        return self.label



import logging

import requests
import validators
from requests.exceptions import ConnectTimeout, ReadTimeout, ConnectionError

from w.services.abstract_service import AbstractService
from w.services.technical.models.request_response import RequestResponse
from w.services.technical.models.request_session import RequestSession

logger = logging.getLogger(__name__)


class RequestService(AbstractService):
    """Service Request"""

    @classmethod
    def init_session(cls, base_url, **options) -> RequestSession:
        """
        Initialize session

        Args:
            base_url (str): base url for request
            **options:
                auth (tuple): credentials (<username>, <password>)
                headers (dict): headers
                cookies (dict): cookies
                timeout (float): number of seconds before timeout

        Returns:
            RequestSession
        """
        if not validators.url(base_url):
            raise RuntimeError("Invalid url '{}'".format(base_url))

        default_options = {"headers": None, "cookies": None}
        options = {**default_options, **options}

        return RequestSession(base_url=base_url, **options)

    @staticmethod
    def _service_unavailable(exception):
        logger.warning(f"Request {exception}")
        return RequestResponse(
            success=False, content="Service unavailable", status_code=503
        )

    @classmethod
    def get_cookies(cls, session: RequestSession):
        return session.request.cookies.get_dict()

    @classmethod
    def get(cls, url, session: RequestSession = None, **kwargs) -> RequestResponse:
        """
        Sends a GET request.
        Args:
            url (str): full url if no session else complete url part
            session (RequestSession|None): session from RequestService::init_session
            **kwargs:  Optional arguments that lib ``request`` takes (like timeout).

        Returns:
            Response
        """

        logger.info('Request GET %s kwargs="%s"', url, kwargs)
        try:
            if session is not None:
                if session.timeout is not None:
                    kwargs["timeout"] = session.timeout
                response = session.request.get(session.render_url(url), **kwargs)
            else:
                response = requests.get(url, **kwargs)
            return RequestResponse(response)
        except (ConnectTimeout, ReadTimeout, ConnectionError) as e:
            return cls._service_unavailable(e)

    @classmethod
    def post(cls, url, data, session: RequestSession = None, **kwargs):
        """
        Sends a POST request.
        Args:
            url (str): full url if no session else complete url part
            data (dict|str): data to post
            session (RequestSession|None): session from RequestService::init_session
            **kwargs:  Optional arguments that lib ``request`` takes (like timeout).

        Returns:
            Response
        """
        logger.info('Request POST %s data="%s" kwargs="%s"', url, data, kwargs)
        try:
            if session is not None:
                if session.timeout is not None:
                    kwargs["timeout"] = session.timeout
                response = session.request.post(
                    session.render_url(url), data=data, **kwargs
                )
            else:
                response = requests.post(url, data=data, **kwargs)
            return RequestResponse(response)
        except (ConnectTimeout, ReadTimeout, ConnectionError) as e:
            return cls._service_unavailable(e)



from dataclasses import asdict, fields


class DataclassMixin:
    def to_dict(self):
        return asdict(self)  # noqa

    def __str__(self):
        return str(self.to_dict())

    @classmethod
    def list_fields(cls):
        return [f.name for f in fields(cls)]  # noqa



import pytest
from django.conf import settings
from django.test.utils import override_settings
from rest_framework.exceptions import AuthenticationFailed

from w.services.technical.models.sso import SsoValidToken
from w.services.technical.sso_service import SsoService
from w.tests.helpers import sso_test_helper
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestSsoService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.user_data = {
            "email": "user@email.com",
            "username": "user@email.com",
            "first_name": "user-firstname",
            "last_name": "user-lastname",
            "password": "user-pwd",
            "client_roles": {"app1": ["role_app1"]},
        }

    def setup_method(self):
        SsoService.clear()
        SsoService.init()
        with sso_test_helper.mock_keycloak_admin_init() as m:
            SsoService._init_keycloak_admin()
        self.keycloak_admin = m

    """
    is_token_valid
    """

    def test_token_is_valid_with_not_initialize_service_raise_runtime_error(self):
        """Ensure method raise RuntimeError if service is not initialized"""
        SsoService.clear()
        match = "Service SsoService must be initialized first"
        with pytest.raises(RuntimeError, match=match):
            SsoService.is_token_valid("need-to-be-initialized")

    def test_token_is_valid_with_invalid_token_raise_error(self):
        """Ensure method raise AuthenticationFailed if token is invalid"""
        match = "Invalid or expired token."
        with sso_test_helper.valid_token_failure():
            with pytest.raises(AuthenticationFailed, match=match):
                SsoService.is_token_valid("invalid-token")

    def test_token_is_valid_with_success_return_sso_token(self):
        """Ensure method succeed"""
        with sso_test_helper.sso_introspect_success() as m:
            actual = SsoService.is_token_valid("valid-token")
        assert isinstance(actual, SsoValidToken)
        self.assert_equals_resultset(
            {"actual": actual.to_dict(), "mock_calls": self.get_mock_calls(m)}
        )

    """
    get_clients
    """

    def test_get_clients_with_not_init_service_raise_error(self):
        """Ensure method raise RuntimeError if not initialized"""
        SsoService.clear()
        match = "Service SsoService must be initialized first"
        with pytest.raises(RuntimeError, match=match):
            SsoService.get_clients()

    def test_get_clients_with_missing_admin_raise_error(self):
        """Ensure method raise RuntimeError if not initialized"""
        SsoService.clear()
        sso_settings = {**settings.SSO}
        sso_settings.pop("ADMIN_PASSWORD")
        SsoService.init()

        match = "Missing SSO settings ADMIN_LOGIN or ADMIN_PASSWORD"
        with override_settings(SSO=sso_settings):
            with pytest.raises(RuntimeError, match=match):
                SsoService.get_clients()

    def test_get_clients_with_success_return_dict(self):
        """Ensure method succeeds"""
        self.keycloak_admin.get_clients.return_value = self.get_dataset(
            "sso/get_clients_success.json"
        )
        actual = SsoService.get_clients()
        self.assert_equals_resultset(actual)

    """
    get_client_roles
    """

    def test_get_client_roles_with_success_return_dict(self):
        """Ensure method succeed"""
        self.keycloak_admin.get_client_roles.return_value = self.get_dataset(
            "sso/get_client_roles.json"
        )
        actual = SsoService.get_client_roles("client-id")
        self.assert_equals_resultset(
            {"actual": actual, "mocks": self.get_mock_calls(self.keycloak_admin)}
        )

    """
    get_or_create_user
    """

    def test_get_or_create_user_with_unknown_user_return_user(self):
        """Ensure method create user in sso server if not exist"""
        self.keycloak_admin.get_user_id.return_value = None
        self.keycloak_admin.create_user.return_value = "user-id"
        self.keycloak_admin.get_clients.return_value = self.get_dataset(
            "sso/get_clients_success.json"
        )
        self.keycloak_admin.get_client_roles.return_value = self.get_dataset(
            "sso/get_client_roles.json"
        )
        self.keycloak_admin.assign_client_role.return_value = None
        self.keycloak_admin.get_user.return_value = self.get_dataset(
            "sso/get_user_success.json"
        )
        actual = SsoService.get_or_create_user(self.user_data)
        self.assert_equals_resultset(
            {"actual": actual, "mocks": self.get_mock_calls(self.keycloak_admin)}
        )
        self.save_as_dataset(
            "sso_service/get_or_create_user_with_unknown_user_return_user.json", actual
        )

    def test_get_or_create_user_with_known_user_return_user(self):
        """Ensure method only retrieve user in sso server if exist"""
        self.keycloak_admin.get_user_id.return_value = "user-id"
        self.keycloak_admin.get_user.return_value = self.get_dataset(
            "sso/get_user_success.json"
        )
        actual = SsoService.get_or_create_user(self.user_data)
        self.assert_equals_resultset(
            {"actual": actual, "mocks": self.get_mock_calls(self.keycloak_admin)}
        )



"""
Need extra django settings configuration

SSO={
    "SERVER_URL": <sso url>,
    "REALM": <sso api realm>,
    "CLIENT_ID": <sso client id api>,
    "CLIENT_SECRET_KEY": <sso client secret see Client credentials>,
    "ADMIN_LOGIN": <optional: sso admin user>
    "ADMIN_PASSWORD": <optional: sso admin password>
}
"""
from typing import Optional

from django.conf import settings
from keycloak import KeycloakOpenID, KeycloakAdmin
from rest_framework.exceptions import AuthenticationFailed

from w.services.abstract_service import AbstractService
from django.utils.translation import gettext as _

from w.services.technical.models.sso import SsoValidToken


class SsoService(AbstractService):
    _keycloak_session: KeycloakOpenID = None
    _keycloak_admin: KeycloakAdmin = None
    _clients: Optional[dict] = None
    _client_roles: dict = {}

    @classmethod
    def clear(cls):
        cls._keycloak_session = None
        cls._keycloak_admin = None
        cls._clients = None
        cls._client_roles = {}

    @classmethod
    def _is_initialized(cls):
        return cls._keycloak_session is not None

    @classmethod
    def _init_keycloak_admin(cls):
        if "ADMIN_LOGIN" in settings.SSO and "ADMIN_PASSWORD" in settings.SSO:
            cls._keycloak_admin = cls._get_keycloak_admin()
            return cls._keycloak_admin

        raise RuntimeError("Missing SSO settings ADMIN_LOGIN or ADMIN_PASSWORD")

    @classmethod
    def _check_admin_is_initialized(cls):
        cls._check_is_initialized()

        if cls._keycloak_admin is not None:
            return cls._keycloak_admin

        return cls._init_keycloak_admin()

    @classmethod
    def _get_keycloak_admin(cls):  # pragma: no cover
        """Only to simplify tests"""
        kc = KeycloakAdmin(
            server_url=settings.SSO["SERVER_URL"],
            username=settings.SSO["ADMIN_LOGIN"],
            password=settings.SSO["ADMIN_PASSWORD"],
            realm_name="master",
        )
        kc.realm_name = settings.SSO["REALM"]
        return kc

    @classmethod
    def init(cls) -> None:
        """Initialize service
        config(dict): Sso configuration:
            server_url(str): sso server url
            realm_name(str): realm name
            client_id(str): client id
            client_secret_key(str): client secret
        """
        config = {
            "server_url": settings.SSO["SERVER_URL"],
            "realm_name": settings.SSO["REALM"],
            "client_id": settings.SSO["CLIENT_ID"],
            "client_secret_key": settings.SSO["CLIENT_SECRET_KEY"],
        }
        cls._keycloak_session = KeycloakOpenID(**config)

    @classmethod
    def _sso_introspect(cls, token):  # pragma: no cover
        return cls._keycloak_session.introspect(token)

    @classmethod
    def get_clients(cls):
        # need client role: realm-management:manage-users
        if cls._clients is None:
            cls._check_admin_is_initialized()
            clients = cls._keycloak_admin.get_clients()
            cls._clients = {c["clientId"]: c["id"] for c in clients}
        return cls._clients

    @classmethod
    def is_token_valid(cls, token):
        cls._check_is_initialized()
        decoded_token = cls._sso_introspect(token)
        if decoded_token.get("active", False):
            return SsoValidToken(decoded_token)
        raise AuthenticationFailed(_("Invalid or expired token."))

    @classmethod
    def create_user(cls, user: dict):
        # need client role: realm-management:manage-users
        return cls._keycloak_admin.create_user(
            {
                "email": user["email"],
                "username": user["username"],
                "enabled": True,
                "firstName": user["first_name"],
                "lastName": user["last_name"],
                "credentials": [{"value": user["password"], "type": "password"}],
            }
        )

    @classmethod
    def add_client_role(cls, user_id, client_id, roles):
        res = cls._keycloak_admin.assign_client_role(user_id, client_id, roles)
        return res

    @classmethod
    def get_client_roles(cls, client_id):
        if client_id not in cls._client_roles:
            cls._check_admin_is_initialized()
            cls._client_roles[client_id] = {
                r["name"]: r for r in cls._keycloak_admin.get_client_roles(client_id)
            }
        return cls._client_roles[client_id]

    @classmethod
    def _add_client_roles(cls, user_id, client_roles):
        clients = cls.get_clients()
        for client, roles in client_roles.items():
            client_id = clients[client]
            client_roles = cls.get_client_roles(client_id)
            roles = [client_roles[r] for r in roles]
            cls.add_client_role(user_id, client_id, roles)

    @classmethod
    def get_or_create_user(cls, user: dict):
        """
        Args:
                "email": "example@example.com",
                "username": "example@example.com",
                "first_name": "Example",
                "last_name": "Example",
                "password": "astrongpassword",
                "client_roles": {
                    "<clientId>": ["role1", ...],
                    "<clientId>": ["roles"]
                    ...
                }
            }

        Need Keycloak real-management roles:
            - manage-users
            - views-client
        """
        cls._check_admin_is_initialized()
        user_id = cls._keycloak_admin.get_user_id(user["username"])
        if not user_id:
            user_id = cls.create_user(user)
            cls._add_client_roles(user_id, user["client_roles"])

        user = cls._keycloak_admin.get_user(user_id)
        return user
